package de.fzj.unicore.bes.security;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

import org.apache.log4j.Logger;

import de.fzj.unicore.uas.util.LogUtil;
import eu.unicore.security.AuthorisationException;



/**
 *
 * The class loads the username/password mapping file by using 
 * the specified path via constructor or from the default location. 
 * This class also matches the incoming username/password elements 
 * from the file-mapped ones.   
 * 
 * @author m.memon
 *
 */
public class FileBasedUsernameValidator {

	private static final Logger logger = LogUtil.getLogger(LogUtil.SECURITY,FileBasedUsernameValidator.class); 

	public static final String USERNAME_MAP_FILE = "username.token.security";

	private static FileBasedUsernameValidator fbInstance = null;

	private final File usersFile;

	private long lastChanged=0;

	private Properties map; 

	private FileBasedUsernameValidator(String file) throws Exception{
		if(file == null){
			throw new IllegalArgumentException();
		}
		logger.info("Using username/password map file: "+file);
		usersFile = new File(file);
		if(!usersFile.exists())	throw new FileNotFoundException("Unable to find the username/password mapping file: "+file);
		update();
	}

	public static synchronized FileBasedUsernameValidator getInstance(String fileName) throws Exception{
		if(fbInstance == null){
			return new FileBasedUsernameValidator(fileName);
		}
		return fbInstance;
	}

	/**
	 * Validates username/password using the mapping file.
         * 
         * Before the validation, the mapping file is reread.
	 * 
	 * @param userName - user name to be validated
	 * @param password - password to be validated
	 * @return boolean true if the user is authorized, false otherwise
         * @throws AuthorisationException if validation fails due to an internal error (IO, etc)
	 * 
	 */
	public synchronized boolean validate(String userName, String password){
		logger.debug("Validating username data");
		try{
			update();
			String p = map.getProperty(userName);
			return p!=null && p.equals(password);
		}
		catch(Exception e){
			throw new AuthorisationException("username authoriser failed",e);
		}
	}

	/*
	 * re-reads the properties
	 */
	private void update()throws IOException{
		if(lastChanged==0 || usersFile.lastModified()>lastChanged){
			map=new Properties();
                        InputStream is = new FileInputStream(usersFile);
			map.loadFromXML(is);
                        is.close();
			lastChanged=usersFile.lastModified();
		}
	}

}
