/*********************************************************************************
 * Copyright (c) 2006-2007 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
 
package de.fzj.unicore.uas.discovery;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import eu.unicore.security.util.Log;

/**
 * Use UDP multicast to find components
 * 
 * @author schuller
 */
public abstract class ComponentFinder implements Runnable{

	protected static final Logger logger=Log.getLogger(Log.UNICORE, ComponentFinder.class);

	private String componentName;
	
	private InetAddress bgroup;
	
	private int port;
	
	protected String key;
	
	private List<String> results=null;
	
	public List<String> getResults() {
		return results;
	}

	public String getComponentName() {
		return componentName;
	}

	protected void doDiscover() {
		List<String> available = discover();
		if(available.size() > 0)
		{
			String reply = choose(available);
			makeResult(reply);
		}
		else
		{
			logger.warn("Nothing found.");
		}
	}
	
	public void run() {
		List<String> results=discover();
		String reply=choose(results);
		makeResult(reply);
		this.results=results;
	}
	
	/**
	 * choose from a list of replies
	 * @param list
	 * @return selected reply
	 */
	protected String choose(List<String>replies){
		if(replies.size()>0)return replies.get(0);
		return null;
	}
	
	protected abstract void makeResult(String reply);
	
	private List<String> discover() {
		try {
			logger.debug("Searching...");
			port = 7700;
			bgroup = InetAddress.getByName("228.5.6.7");
			int timeout = 5000;
			return lookup(getMessage(), timeout);
		} catch (UnknownHostException ex) {
			Log.logException("",ex,logger);
			return null;
		}

	}

	protected void sendRequest(String message) {
	      try {
	    	    String request = "Request:"+key+":"+message;
	            logger.debug("sending "+request);
	            MulticastSocket s = new MulticastSocket(port);
	            s.joinGroup(bgroup);
	            DatagramPacket hi = new DatagramPacket(request.getBytes(),
	                    request.length(),
	                    bgroup, port);
	            s.send(hi);
	        } catch (Exception ex) {
	            Log.logException("Could not send multicast request: ",ex,logger);
	        }
	    }

    /** 
     * this collects responses until <i>timeout</i> msec is reached
     */
    private List<String> lookup(String message, int timeout) {
        List<String> results = new ArrayList<String>();
        sendRequest(message);
        String result = null;
        long startingTime = System.currentTimeMillis();
        try
        {
        	do{
        		result = waitForAnswer(timeout);
        		if(result != null && !results.contains(result)){
        			try{
        				results.add(result);
        				break;
        			}
        			catch(Exception e){
        				logger.warn("Received message but could not parse: "+result);
        			}
        		}
        		if(timeout>0 && System.currentTimeMillis()-startingTime > timeout)
        			throw new IOException();
        	} while (true);
        } catch (IOException e) {
        	if(results.size() < 1) logger.debug("Discovery timed out.");
        }
        return results;
    }
    
    private String waitForAnswer(int timeout) throws IOException{
    	MulticastSocket s2 = new MulticastSocket(port);
    	s2.joinGroup(bgroup);
    	s2.setSoTimeout(timeout); // timout: 5 sec.
    	byte[] buf = new byte[1000];
    	DatagramPacket recv = new DatagramPacket(buf, buf.length);
    	s2.receive(recv);
    	String response = (new String(recv.getData())).trim();
    	if(response.startsWith("Response:"+key)) {
    		logger.debug("Received: "+response+ " from "+recv.getAddress());
    		return response.replaceFirst("Response:"+key+":","");

    	} else return null;
    }
   
   /**
    * the additional info to send to the server
    */
   protected abstract String getMessage();

}
