/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/

package de.fzj.unicore.uas;

import java.util.Calendar;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.unigrids.services.atomic.types.ProtocolType;

import de.fzj.unicore.uas.admin.UASAdmin;
import de.fzj.unicore.uas.impl.registry.RegistryHandler;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.Home;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.impl.DefaultHome;
import de.fzj.unicore.wsrflite.utils.deployment.PropertyChecker;
import de.fzj.unicore.wsrflite.xmlbeans.sg.ServiceGroupEntry;

/**
 * main class, intended to run the UAS as a set of standalone services,
 * including a target system factory, job management etc. It is based on the
 * WSRFlite hosting environment, and uses the XNJS as back-end execution engine.
 * 
 * @author schuller
 */
public class UAS implements PropertyChecker {

	private static final Logger logger = LogUtil.getLogger(LogUtil.UNICORE,
			UAS.class);

	public static final String VERSION = UAS.class.getPackage()
			.getImplementationVersion();

	/**
	 * for unit testing, set this property to "true" to disable security
	 */
	public static final String DISABLE_SECURITY_AND_ACCESS_CONTROL = "uas.security.disable";

	/**
	 * Enable registration with a gateway
	 */
	public static final String AUTOREGISTER_WITH_GATEWAY_KEY = "uas.gatewayregistration";

	/**
	 * update interval for the registration with a gateway in seconds (default =
	 * 30 second)
	 */
	public static final String AUTOREGISTER_WITH_GATEWAY_UPDATE_KEY = "uas.gatewayregistration.updateinterval";

	/**
	 * Enable multicast discovery of a gateway
	 */
	public static final String DISCOVER_GATEWAY_KEY = "uas.gatewaydiscovery";

	/**
	 * To override the default UAS startup code, set this to a list of names
	 * (space-separated) of classes implementing java.lang.Runnable
	 */
	public static final String ON_STARTUP_KEY = "uas.onstartup";

	/**
	 * whether to run a connections check after startup (defaults to "true")
	 */
	public static final String ON_STARTUP_SELFTEST = "uas.onstartup.selftest";

	/**
	 * To enable waiting for the gateway on startup, set this property to 'true'
	 */
	public static final String ON_STARTUP_WAIT_KEY = "uas.onstartup.wait";

	/**
	 * To enable waiting for the gateway on startup, set this property to the
	 * max wait time in seconds default: 120
	 */
	public static final String ON_STARTUP_WAITTIME_KEY = "uas.onstartup.wait.time";

	/**
	 * property defining the vsite name (i.e. default TSS name)
	 */
	public static final String TSS_NAME_PROPERTY = "uas.targetsystem.sitename";

	/**
	 * property defining whether to disable users' home SMSes
	 */
	public static final String TSS_DISABLE_HOME_PROPERTY = "uas.targetsystem.home.disable";

	/**
	 * if set to <code>true</code>, the storages attached to a TSS will always
	 * have unique IDs. In the default case, storage names will be formed from
	 * the user's xlogin and the storage name, e.g. "a.user-Home".
	 */
	public static final String TSS_FORCE_UNIQUE_STORAGE_IDS = "uas.targetsystem.storage.force_unique_ids";

	/**
	 * Property defining the "default" protocols supported by the SMS<br/> 
	 * This is a list of String-separated protocol names, e.g. "RBYTEIO SBYTEIO"
	 */
	public static final String SMS_PROTOCOLS = "uas.sms.protocols";

	/**
	 * When doing file transfers, UNICORE tries to detect whether two storage
	 * resources are accessing the same filesystem. If yes, the transfer is done
	 * by direct copying. Set to "true" to disable this feature.
	 */
	public final static String SMS_TRANSFER_FORCEREMOTE = "uas.sms.filetransfer.forceremote";

	/**
	 * the name of the WSRFlite config file
	 */
	public static final String WSRFLITE_CONFIG = "uas.wsrflite.configfile";

	/**
	 * set the name or the XNJS config file
	 */
	public static final String TSF_XNJS_CONFIGFILE = "uas.targetsystemfactory.xnjs.configfile";

	// service names

	public static final String TSF = "TargetSystemFactoryService";
	public static final String TSS = "TargetSystemService";
	public static final String JMS = "JobManagement";
	public static final String SMS = "StorageManagement";
	public static final String SMF = "StorageFactory";
	public static final String REG = "Registry";
	public static final String REGENTRY = ServiceGroupEntry.SERVICENAME;
	public static final String RESERVATIONS = "ReservationManagement";
	public static final String ENUMERATION = "Enumeration";
	public static final String META = "MetadataManagement";
	public static final String TASK = "Task";
	
	// registry related
	public static final String GLOBAL_REGISTRY_ADVERTISE = "uas.registry.global.advertise";
	public static final String EXTERNAL_REGISTRY_USE = "uas.externalregistry.use";
	public static final String REGISTRY_AUTODISCOVER = "uas.externalregistry.autodiscover";
	public static final String EXTERNAL_REGISTRY_KEY = "uas.externalregistry.url";

	/**
	 * you need to add the protocol to this one to get the right service e.g.
	 * FTS_BASE+"RBYTEIO" = "FileTransferRBYTEIO"
	 */
	public static final String FTS_BASE = "FileTransfer";
	
	/**
	 * If one of the following is used as protocol for file transfers instead of
	 * a {@link ProtocolType} constant, UNICORE should automatically try to
	 * negotiate a suitable file transfer protocol.
	 */
	public static final String[] AUTO_NEGOTIATE_FT_PROTOCOL = {"u6","unicore6","unicore"};

	private static UASAdmin jmxBean;

	public static final String getVersion() {
		String ver = VERSION != null ? VERSION : "DEVELOPMENT";
		return ver;
	}

	public static final String getHeader() {
		return "UNICORE/X Version " + getVersion();
	}

	/**
	 * @deprecated
	 */
	public static Properties getProperties() {
		return Kernel.getKernel().getProperties();
	}

	/**
	 * @deprecated
	 */
	public static String getProperty(String key) {
		return Kernel.getKernel().getProperty(key);
	}

	public final Calendar upSince = Calendar.getInstance();

	
	public String getConnectionStatus(){
		if(!kernel.getBooleanProperty("uas.onstartup.selftest", true))return "";
		StringBuilder report = new StringBuilder();
		String newline = System.getProperty("line.separator");
		report.append(newline);
		report.append("Connection status report");
		report.append(newline);
		report.append("************************");
		report.append(newline);
		report.append("Gateway connection: ");
		report.append(jmxBean.getGatewayConnectionStatus());
		report.append(newline);
	
		try{
			if (!RegistryHandler.get(kernel).isGlobalRegistry()) {
				report.append("Registry connection: ");
				report.append(jmxBean.getRegistryConnectionStatus());
				report.append(newline);
				report.append("User attribute sources: ");
				report.append(jmxBean.getAttributeSourcesStatus());
				report.append(newline);
				if (kernel.getHome(UAS.TSS) != null) {
					report.append("TSI connection: ");
					report.append(jmxBean.getTSIConnectionStatus());
					report.append(newline);
				}
			}
		}catch(Exception ex){
			logger.error(ex);
		}
		return report.toString();
	}

	private final String configFile;

	private final Kernel kernel;
	
	/**
	 * @param configFile
	 */
	public UAS(String configFile) throws Exception {
		System.out.println("Reading config from " + configFile);
		jmxBean=new UASAdmin(this);
		Kernel.addMBean(jmxBean, "UNICORE/X");
		this.configFile=configFile;
		this.kernel=new Kernel(); 
	}

	public Kernel getKernel(){
		return kernel;
	}

	public void start()throws Exception{
		kernel.start(configFile);
		printInfo();
	}
	
	public void startSynchronous()throws Exception{
		kernel.setPropertyChecker(this);
		kernel.startSynchronous(configFile);
		printInfo();
	}
	
	private void printInfo(){
		String msg=getHeader()+getConnectionStatus();
		System.out.println(msg);
		logger.info(msg);
	}
	/**
	 * start the UAS<br/>
	 * 
	 * args: optional config file name
	 * 
	 */
	public static void main(String[] args) throws Exception {
		try{
			UAS uas=new UAS(args[0]);
			uas.startSynchronous();
			uas.forceRefreshRegistryEntries();
		}catch(Exception ex){
			ex.printStackTrace();
			System.err.println("ERROR during server startup, server NOT started.");
			System.exit(1);
		}
	}

	private void forceRefreshRegistryEntries() {
		try {
			if (RegistryHandler.get(kernel).isGlobalRegistry())
				return;
			
			Home home = kernel.getHome(ServiceGroupEntry.SERVICENAME);
			if (home != null && home instanceof DefaultHome) {
				((DefaultHome) home).runExpiryCheckNow();
			}
		} catch (Exception ex) {
			LogUtil.logException("Error running expiry checks for <"
					+ ServiceGroupEntry.SERVICENAME + ">", ex, logger);
		}
	}

	/**
	 * check properties and replace obsolete values
	 */
	public void checkProperties(Properties properties, Logger log){
		//check removed UNICORE filetransfer implementations
		String bft=getProperty("uas.filetransfer.protocol.BFT");
		if ("de.fzj.unicore.uas.fts.BaselineFileTransferImpl".equals(bft)) {
			// fallback to default
			bft = "de.fzj.unicore.uas.fts.http.HttpFileTransferImpl";
			log.debug("Property '"
					+ "uas.filetransfer.protocol.BFT"
					+ "' is set to obsolete value, using updated value <"
					+ bft + ">");
			properties.put("uas.filetransfer.protocol.BFT", bft);
		}

	}
	
	public String checkProperty(String key, String value, Logger log){
		if("uas.filetransfer.protocol.BFT".equals(key) &&
				"de.fzj.unicore.uas.fts.BaselineFileTransferImpl".equals(value)){
			String bft = "de.fzj.unicore.uas.fts.http.HttpFileTransferImpl";
			log.debug("Property '"
					+ "uas.filetransfer.protocol.BFT"
					+ "' is set to obsolete value, using updated value <"
					+ bft + ">");
			return bft;
		}
	
		return value;
	}
	
	
}
