/*
 * Copyright (c) 2010 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licensing information.
 *
 * Author: T. Rękawek <newton@mat.uni.torun.pl>
 */

package de.fzj.unicore.uas.impl.dsms;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;
import org.unigrids.services.atomic.types.GridFileType;
import org.unigrids.services.atomic.types.PermissionsType;
import org.unigrids.services.atomic.types.ProtocolType;
import org.unigrids.x2006.x04.services.sms.ChangePermissionsDocument;
import org.unigrids.x2006.x04.services.sms.ChangePermissionsResponseDocument;
import org.unigrids.x2006.x04.services.sms.CopyDocument;
import org.unigrids.x2006.x04.services.sms.CopyResponseDocument;
import org.unigrids.x2006.x04.services.sms.CreateDirectoryDocument;
import org.unigrids.x2006.x04.services.sms.CreateDirectoryResponseDocument;
import org.unigrids.x2006.x04.services.sms.DeleteDocument;
import org.unigrids.x2006.x04.services.sms.DeleteResponseDocument;
import org.unigrids.x2006.x04.services.sms.ExportFileDocument;
import org.unigrids.x2006.x04.services.sms.ExportFileResponseDocument;
import org.unigrids.x2006.x04.services.sms.FindDocument;
import org.unigrids.x2006.x04.services.sms.FindDocument.Find;
import org.unigrids.x2006.x04.services.sms.FindResponseDocument;
import org.unigrids.x2006.x04.services.sms.FindResponseDocument.FindResponse;
import org.unigrids.x2006.x04.services.sms.ImportFileDocument;
import org.unigrids.x2006.x04.services.sms.ImportFileResponseDocument;
import org.unigrids.x2006.x04.services.sms.ListDirectoryDocument;
import org.unigrids.x2006.x04.services.sms.ListDirectoryResponseDocument;
import org.unigrids.x2006.x04.services.sms.ListDirectoryResponseDocument.ListDirectoryResponse;
import org.unigrids.x2006.x04.services.sms.ListPropertiesDocument;
import org.unigrids.x2006.x04.services.sms.ListPropertiesResponseDocument;
import org.unigrids.x2006.x04.services.sms.ReceiveFileDocument;
import org.unigrids.x2006.x04.services.sms.ReceiveFileResponseDocument;
import org.unigrids.x2006.x04.services.sms.RenameDocument;
import org.unigrids.x2006.x04.services.sms.RenameResponseDocument;
import org.unigrids.x2006.x04.services.sms.SendFileDocument;
import org.unigrids.x2006.x04.services.sms.SendFileResponseDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import pl.edu.icm.dsms.catalogue.CatalogueClient;
import pl.edu.icm.dsms.catalogue.CatalogueClient.LdasAndEprs;
import pl.edu.icm.dsms.catalogue.CatalogueUtils;
import pl.edu.icm.dsms.catalogue.exception.FileNotExistsFault;
import pl.edu.icm.dsms.catalogue.model.Lda;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.TransferControllerClient;
import de.fzj.unicore.uas.impl.dsms.Utils.LdaWithStorageClient;
import de.fzj.unicore.uas.impl.dsms.Utils.PhysicalFilenameAndClient;
import de.fzj.unicore.uas.impl.sms.SMSBaseImpl;
import de.fzj.unicore.uas.metadata.MetadataManager;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.ThreadingServices;
import de.fzj.unicore.wsrflite.messaging.ResourceDeletedMessage;
import de.fzj.unicore.wsrflite.persistence.Persist;
import de.fzj.unicore.wsrflite.security.ETDAssertionForwarding;
import de.fzj.unicore.wsrflite.security.ISecurityProperties;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;
import de.fzj.unicore.xnjs.io.Permissions;
import eu.unicore.security.Client;
import eu.unicore.security.SecurityTokens;
import eu.unicore.security.util.Log;

/**
 * Distributed Storage Management Service
 * @author newton
 *
 */
public class DSMSImpl extends SMSBaseImpl
{
	private static final Logger logger = Log.getLogger(LogUtil.SERVICES, DSMSImpl.class);
	
	//default protocol for copy operations
	private final static String protocol = "BFT";
	
	public static final String DSMS = "DSms";
	
	public static final String GET_FREE_SPACE = "uas.dsms.get_free_space";
	public static final String ENABLE_SYNC_1 = "uas.dsms.sync1";
	public static final String ENABLE_SYNC_2 = "uas.dsms.sync2";
//	public static final String IGNORE_COPY_FAULT = "uas.dsms.ignore_copy_fault";

	@Persist
	private String dSmsId;
	
	@Persist
	private EndpointReferenceType catalogueEpr;
		
	@Override
	public void initialise(String name, Map<String,Object>initobjs) throws Exception
	{
		super.initialise(name, initobjs);
		properties.put(RPFileSystem, new DSMSFileSystemRP(this, "DSMS"));
		workdir = getStorageRoot();
		
		if(storageFactoryID != null)
			dSmsId = getCatalogueClient().createSms();
		logger.info("Initialised dSMS " + (dSmsId == null ? "" : dSmsId));
	}

	/**
	 * metadata management is disabled
	 * TODO handle distributed metadata
	 */
	protected MetadataManager getMetadataManager()throws Exception{
		return null;
	}
	
	@Override
	public void destroy()
	{
		if(storageFactoryID != null)
		{
			try
			{
				ResourceDeletedMessage m=new ResourceDeletedMessage("deleted:" + getUniqueID());
				m.setDeletedResource(getUniqueID());
				m.setServiceName(getServiceName());
				getKernel().getMessaging().getChannel(storageFactoryID).publish(m);
			}
			catch(Exception e)
			{
				logger.error("Could not send message to StorageFactory", e);
			}
		}
		try
		{
			if(cleanupOnDestroy != null && cleanupOnDestroy)
			{
				LdasAndEprs lae = getCatalogueClient().removeSms(dSmsId);
				for(EndpointReferenceType epr : lae.getEprs())
				{
					ISecurityProperties sec = getKernel().getSecurityProperties().clone();
					SecurityTokens tokens = getSecurityTokens();
					Client client = getKernel().getSecurityManager().createAndAuthoriseClient(tokens);
					ETDAssertionForwarding.configureETD(client, sec);
					StorageClient storageClient = new StorageClient(epr, sec);
					storageClient.destroy();
				}
				for(Lda lda : lae.getLdas())
				{
					StorageClient client = Utils.getStorageClient(lda, getClient(),kernel);
					client.delete(lda.getPhysicalName());
				}
			}
		}
		catch(Exception e)
		{
			logger.error("Could not destroy dSMS", e);
		}
		super.destroy();
		logger.info("Destroyed dSMS " + (dSmsId == null ? "" : dSmsId));
	}

	
	public ChangePermissionsResponseDocument ChangePermissions(ChangePermissionsDocument in) throws BaseFault
	{
		String file = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			file = makeSMSLocal(in.getChangePermissions().getPath());
			PermissionsType p = in.getChangePermissions().getPermissions();
			Permissions permissions = new Permissions(p.getReadable(), p.getWritable(), p.getExecutable());
			
			LdaWithStorageClient lda = Utils.syncLookup(catalogue, file, dSmsId, getClient(),kernel);
			lda.getStorage().changePermissions(lda.getLda().getPhysicalName(), p.getReadable(),
					p.getWritable(), p.getExecutable());
						
			logger.debug("Changed user permissions for '"+file+"' to <"+permissions.toString()+">");

			ChangePermissionsResponseDocument res = ChangePermissionsResponseDocument.Factory.newInstance();
			res.addNewChangePermissionsResponse();
			return res;
		}
		catch(Exception e)
		{
			LogUtil.logException("Could not change permissions for file "+ file, e);
			String msg=LogUtil.createFaultMessage("Could not change permissions for file "+ file, e);
			throw BaseFault.createFault(msg);
		}
	}

	public CopyResponseDocument Copy(CopyDocument in) throws BaseFault
	{
		String source = "", target = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			source = makeSMSLocal(in.getCopy().getSource());
			target = makeSMSLocal(in.getCopy().getDestination());

			LdaWithStorageClient lSource = Utils.syncLookup(catalogue, source, dSmsId, getClient(),kernel);
			PhysicalFilenameAndClient lTarget = Utils.addFile(target, dSmsId, catalogue, getClient(),kernel);
			
			TransferControllerClient c;
			c = lSource.getStorage().sendFile(lSource.getLda().getPhysicalName(), Utils.getUrl(lTarget.getClient().getEPR(), lTarget.getPhysicalFilename(), protocol));
			
			//wait just long enough to make sure the transfer was kicked off...
			int time=0;
			while(!c.isComplete() && !c.hasFailed() && time<5)
			{
				Thread.sleep(1000);
				time++;
			}
			if(c.hasFailed())
			{
				String err=c.getStatus();
				throw new Exception("Send file failed: "+err);
			}
			
			logger.debug("Copied '"+source+"' -> '"+target+"'");
			
			CopyResponseDocument res=CopyResponseDocument.Factory.newInstance();
			res.addNewCopyResponse();
			return res;
		}
		catch(Exception e)
		{
			String msg=LogUtil.createFaultMessage("Could not copy " + source + " to " + target, e);
			LogUtil.logException("Could not copy " + source + " to " + target, e);
			throw BaseFault.createFault(msg);
		}

	}

	public CreateDirectoryResponseDocument CreateDirectory(CreateDirectoryDocument in) throws BaseFault
	{
		String path = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();	
			
			path = makeSMSLocal(in.getCreateDirectory().getPath());
			catalogue.addDir(path, dSmsId);
			
			// in order to create dir on physical sms
			String tempFile = path + "/tempFile";
			Utils.addFile(tempFile, dSmsId, catalogue, getClient(),kernel);
			catalogue.remove(tempFile, dSmsId);
			
			logger.debug("Created directory " + path);
		}
		catch(Exception e)
		{
			String msg=Log.createFaultMessage("Could not create directory " + path, e);
			Log.logException("Could not create directory " + path, e,logger);
			throw BaseFault.createFault(msg);
		}
		CreateDirectoryResponseDocument res=CreateDirectoryResponseDocument.Factory.newInstance();
		res.addNewCreateDirectoryResponse();
		return res;
	}

	public DeleteResponseDocument Delete(DeleteDocument in) throws BaseFault
	{
		String path = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			path = makeSMSLocal(in.getDelete().getPath());
			if(catalogue.isDir(path, dSmsId))
			{
				List<Lda> ldaList = catalogue.removeDir(path, dSmsId);
				for(Lda lda : ldaList)
					Utils.getStorageClient(lda, getClient(),kernel).delete(lda.getPhysicalName());
			}
			else
			{
				Lda lda = catalogue.remove(path, dSmsId);
				Utils.getStorageClient(lda, getClient(),kernel).delete(lda.getPhysicalName());
			}
			
			logger.debug("Deleted '" + path + "'");
			DeleteResponseDocument res=DeleteResponseDocument.Factory.newInstance();
			res.addNewDeleteResponse();
			return res;
		}
		catch(Exception e)
		{
			String msg=Log.createFaultMessage("Could not delete " + path, e);
			Log.logException("Could not delete " + path, e, logger);
			throw BaseFault.createFault(msg);
		}
	}

	public ExportFileResponseDocument ExportFile(ExportFileDocument in) throws BaseFault
	{
		String source = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			source = makeSMSLocal(in.getExportFile().getSource());
			ProtocolType.Enum protocol=in.getExportFile().getProtocol();
			
			LdaWithStorageClient lda = Utils.syncLookup(catalogue, source, dSmsId, getClient(),kernel);
			
			logger.debug("Export: '" + source + "' " + protocol);
			in.getExportFile().setSource(lda.getLda().getPhysicalName());
			return lda.getStorage().ExportFile(in);
		}
		catch (Exception e)
		{
			String msg=Log.createFaultMessage("Could not export "+source,e);
			Log.logException("Could not export " + source, e, logger);
			throw BaseFault.createFault(msg);
		}
	}

	public FindResponseDocument Find(FindDocument in) throws BaseFault
	{
		FindResponseDocument resd = FindResponseDocument.Factory.newInstance();
		FindResponse res = resd.addNewFindResponse();
		String base = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			Find find = in.getFind();
			base = find.getBase();
			logger.debug("Finding in '"+base+"'");
			
			Boolean recurse = find.getRecurse();
			Boolean regexp = false;
			String name = null;
			if(find.getFilter().isSetNameMatchRegExp())
			{
				regexp = true;
				name = find.getFilter().getNameMatchRegExp();
			}
			else
				name = find.getFilter().getNameMatch();
			Calendar before = find.getFilter().getBefore();
			Calendar after = find.getFilter().getAfter();
			
			int i = 0;
			List<Lda> ldas = catalogue.lookupDir(base, dSmsId);
			if(ldas == null || ldas.size() == 0)
				return resd;
			for(Lda lda : ldas)
			{
				StorageClient client;
				GridFileType files[];
				try
				{
					client = Utils.getStorageClient(lda, getClient(),kernel);
					files = client.find(lda.getPhysicalName(), recurse, name, regexp, before, after);
				}
				catch(Exception e)
				{
					if(lda != null)
						logger.error("Can't connect to SMS " + lda.getSmsAddress(), e);
					else
						logger.error("Can't connect to SMS", e);
					continue;
				}
				
				for(GridFileType g : files)
				{
					g.setPath(CatalogueUtils.getLogicalName(g.getPath()));
					res.addNewGridFile();
					res.setGridFileArray(i++, g);
				}
			}
			return resd;
		}
		catch(Exception e)
		{
			String msg=Log.createFaultMessage("Could not perform find operation in base " + base, e);
			Log.logException("Could not perform find operation in base " + base, e, logger);
			throw BaseFault.createFault(msg);
		}
	}

	public ImportFileResponseDocument ImportFile(ImportFileDocument in) throws BaseFault
	{
		String dest = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			dest = makeSMSLocal(in.getImportFile().getDestination());
			logger.debug("Importing '"+dest+"'");
			
			PhysicalFilenameAndClient pfac = Utils.addFile(dest, dSmsId, catalogue, getClient(),kernel);
			
			in.getImportFile().setDestination(pfac.getPhysicalFilename());
			return pfac.getClient().ImportFile(in);
		}
		catch (Exception e)
		{
			String msg=Log.createFaultMessage("Could not import " + dest, e);
			Log.logException("Could not import " + dest, e, logger);
			throw BaseFault.createFault(msg);
		}
	}

	public ListDirectoryResponseDocument ListDirectory(
			ListDirectoryDocument in) throws BaseFault
	{
		String p = "";
		
		ListDirectoryResponseDocument resd = ListDirectoryResponseDocument.Factory.newInstance();
		ListDirectoryResponse res = resd.addNewListDirectoryResponse();
		BigInteger offsetP = in.getListDirectory().getOffset();
		int offset = (offsetP != null ? offsetP.intValue() : 0);
		BigInteger limitP = in.getListDirectory().getLimit();
		int limit = limitP != null ? limitP.intValue() : SMSBaseImpl.MAX_LS_RESULTS;
		if(limit > SMSBaseImpl.MAX_LS_RESULTS)
		{
			String msg = "Could not list directory: the requested number of results " +
			"exceeds the internal limit of <"+SMSBaseImpl.MAX_LS_RESULTS+">. " +
			"Please use the limit and offset parameters!";
			logger.warn(msg);
			throw BaseFault.createFault(msg);
		}
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			p = makeSMSLocal(in.getListDirectory().getPath());
			logger.debug("Listing '"+p+"'");
			
			List<Lda> ldas = catalogue.lookupDir(p, dSmsId);
			if(ldas == null || ldas.size() == 0)
				return resd;
			Collections.sort(ldas, new Comparator<Lda>()
			{
				public int compare(Lda o1, Lda o2)
				{
					int r = o1.getSmsAddress().compareTo(o2.getSmsAddress());
					if(r != 0)
						return r;
					r = o1.getPhysicalName().compareTo(o2.getPhysicalName());
						return r;
				}
			});
			
			ExecutorService executor = ThreadingServices.get(kernel).getExecutorService();
			List<Future<GridFileType[]>> futures = new ArrayList<Future<GridFileType[]>>();
			List<GridFileType[]> fileLists = new ArrayList<GridFileType[]>();

			for(Lda lda : ldas)
				futures.add(executor.submit(new ListDirectoryCallable(lda, getClient(), kernel)));
			
			for(Future<GridFileType[]> future : futures)
				try
				{
					if (future.get() != null)
						fileLists.add(future.get());
				} catch (Exception e)
				{
					logger.error("Can't get file list", e.getCause());
				}
			
			int i = 0, j = 0;
			Set<String> usedNames = new HashSet<String>();
			
			for(GridFileType[] files : fileLists)
			{
				for(GridFileType file : files)
				{
					if(j >= limit)
						break;
					if(usedNames.contains(file.getPath()))
						continue;
					usedNames.add(file.getPath());
					if(i++ >= offset)
					{
						file.setPath(CatalogueUtils.getLogicalName(file.getPath()));
						res.addNewGridFile();
						res.setGridFileArray(j++, file);
					}
				}
				if(j >= limit)
					break;
			}
		}
		catch(Exception e)
		{
			String msg=Log.createFaultMessage("Could not list directory " + p, e);
			Log.logException("Could not list directory " + p, e, logger);
			throw BaseFault.createFault(msg);
		}
		return resd;
	}

	public ListPropertiesResponseDocument ListProperties(ListPropertiesDocument in) throws BaseFault
	{
		ListPropertiesResponseDocument res = ListPropertiesResponseDocument.Factory.newInstance();
		res.addNewListPropertiesResponse();
		String path = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();

			String request = in.getListProperties().getPath();
			path = makeSMSLocal(request);
			logger.debug("Listing properties of " + path);
			
			Lda lda;
			StorageClient storage;
			if(!catalogue.isDir(path, dSmsId))
			{
				LdaWithStorageClient _lda = Utils.syncLookup(catalogue, path, dSmsId, getClient(),kernel);
				lda = _lda.getLda();
				storage = _lda.getStorage();
			}
			else
			{
				lda = catalogue.lookupDir(path, dSmsId).get(0);
				storage = Utils.getStorageClient(lda, getClient(),kernel);
			}
			
			GridFileType file = storage.listProperties(lda.getPhysicalName());
			file.setPath(CatalogueUtils.getLogicalName(file.getPath()));
			res.getListPropertiesResponse().setGridFile(file);
			return res;
		}
		catch(FileNotExistsFault e)
		{
		}
		catch(Exception e)
		{
			String msg=Log.createFaultMessage("Could not list properties of "+path, e);
			Log.logException("Could not list properties of "+path, e, logger);
			throw BaseFault.createFault(msg);
		}
		return res;
	}

	public ReceiveFileResponseDocument ReceiveFile(ReceiveFileDocument in) throws BaseFault
	{
		String sourceURI = in.getReceiveFile().getSource();
		String target = makeSMSLocal(in.getReceiveFile().getDestination());
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			PhysicalFilenameAndClient pfac = Utils.addFile(target, dSmsId, catalogue, getClient(),kernel);
			
			logger.debug("Receiving data from <"+sourceURI+"> target is <"+target+">");
			in.getReceiveFile().setDestination(pfac.getPhysicalFilename());
			return pfac.getClient().ReceiveFile(in);
		}
		catch(Exception e)
		{
			String msg=Log.createFaultMessage("Could not receive file from " + sourceURI, e);
			Log.logException("Could not receive file from " + sourceURI, e, logger);
			throw BaseFault.createFault(msg);
		}
	}

	public RenameResponseDocument Rename(RenameDocument in) throws BaseFault
	{
		String source = "", target = "";
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			source = makeSMSLocal(in.getRename().getSource());
			target = makeSMSLocal(in.getRename().getDestination());

			LdaWithStorageClient ldaSrc = Utils.syncLookup(catalogue, source, dSmsId, getClient(),kernel);
			catalogue.remove(source, dSmsId);
			Lda ldaDst = catalogue.add(target, dSmsId, ldaSrc.getLda().getSms());
			
			StorageClient client = ldaSrc.getStorage();
			Utils.makeParentDirs(ldaDst.getPhysicalName(), client);
			client.rename(ldaSrc.getLda().getPhysicalName(), ldaDst.getPhysicalName());
			
			logger.debug("Renamed '" + source + "' -> '" + target + "'");
			RenameResponseDocument res = RenameResponseDocument.Factory.newInstance();
			res.addNewRenameResponse();
			return res;
		}
		catch(Exception e)
		{
			String msg=Log.createFaultMessage("Could not rename " + source + " to " + target, e);
			Log.logException("Could not rename " + source + " to " + target, e, logger);
			throw BaseFault.createFault(msg);
		}
	}

	public SendFileResponseDocument SendFile(SendFileDocument in) throws BaseFault
	{
		String source = in.getSendFile().getSource();
		String targetURI = in.getSendFile().getDestination();
		try
		{
			CatalogueClient catalogue = getCatalogueClient();
			
			String path = makeSMSLocal(source);
			LdaWithStorageClient lda = Utils.syncLookup(catalogue, path, dSmsId, getClient(),kernel);
			
			logger.debug("Sending data to <"+targetURI+"> source is <"+source+">");
			
			in.getSendFile().setSource(lda.getLda().getPhysicalName());
			return lda.getStorage().SendFile(in);
		}
		catch (Exception e)
		{
			String msg=Log.createFaultMessage("Could not send file to " + targetURI, e);
			Log.logException("Could not send file to " + targetURI, e, logger);
			throw BaseFault.createFault(msg);
		}
	}
	
	@Override
	protected String getStorageRoot()
	{
		return "/tmp";
	}
	
	/**
	 * Get catalogue client. If it's possible, Catalogue connection will be made
	 * on behalf of user.
	 * 
	 * @return
	 * @throws Exception
	 */
	public CatalogueClient getCatalogueClient()	throws Exception
	{
		CatalogueClient cat = null;
		if(catalogueEpr != null)
		{
			cat = Utils.getCatalogueClient(getClient(), catalogueEpr, kernel);
			if(cat != null)
			{
				logger.debug("got dcatalogue from cached epr");
				return cat;
			}
		}
		
		logger.debug("looking for dcataloge in registry");
		
		cat = Utils.getCatalogueClientFromRegistry(getClient(), kernel);
		if(cat != null)
			catalogueEpr = cat.getEPR();
		return cat;
	}
}