/*
 * Copyright (c) 2010 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licensing information.
 *
 * Author: T. Rękawek <newton@mat.uni.torun.pl>
 */

package de.fzj.unicore.uas.impl.dsms;

import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import javax.security.auth.x500.X500Principal;
import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.oasisOpen.docs.wsrf.sg2.EntryType;
import org.unigrids.services.atomic.types.GridFileType;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import pl.edu.icm.dsms.catalogue.Catalogue;
import pl.edu.icm.dsms.catalogue.CatalogueClient;
import pl.edu.icm.dsms.catalogue.CatalogueClient.LdaOrFactory;
import pl.edu.icm.dsms.catalogue.exception.FileNotExistsFault;
import pl.edu.icm.dsms.catalogue.exception.InvalidPathFault;
import pl.edu.icm.dsms.catalogue.model.Lda;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.StorageFactoryClient;
import de.fzj.unicore.uas.impl.registry.RegistryHandler;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.ThreadingServices;
import de.fzj.unicore.wsrflite.security.ETDAssertionForwarding;
import de.fzj.unicore.wsrflite.security.ISecurityProperties;
import de.fzj.unicore.wsrflite.utils.WSServerUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.client.ExternalRegistryClient;
import eu.unicore.security.Client;
import eu.unicore.security.util.Log;
import eu.unicore.security.util.client.HttpUtils;

/**
 * Static methods often used in DSMSImpl.
 * 
 * @author newton
 * 
 */
public class Utils
{
	private static final Logger logger = Log.getLogger(LogUtil.SERVICES,
			Utils.class);

	/**
	 * Add file to Catalogue and return LDA.
	 * 
	 * @param logicalFilename
	 * @param dSmsId
	 * @param catalogue
	 * @param client
	 * @return Object containing StorageClient and physical filename.
	 * @throws Exception
	 */
	public static PhysicalFilenameAndClient addFile(String logicalFilename,
			String dSmsId, CatalogueClient catalogue, Client client, Kernel k)
			throws Exception
	{
		LdaOrFactory ldaOrFactory = null;
		Lda lda = null;
		StorageClient storageClient = null;
		boolean sync1 = isSyncEnabled(k,1);

		try
		{
			LdaWithStorageClient _lda = syncLookup(catalogue, logicalFilename,
					dSmsId, client, sync1, false, k);
			lda = _lda.getLda();
			storageClient = _lda.getStorage();
		} catch (FileNotExistsFault e)
		{
			ldaOrFactory = catalogue.add(logicalFilename, dSmsId);
		}

		String physicalFilename = null;
		if (ldaOrFactory != null && ldaOrFactory.isFactory())
		{
			ISecurityProperties sec = k.getSecurityProperties().clone();
			ETDAssertionForwarding.configureETD(client, sec);

			StorageFactoryClient factClient = new StorageFactoryClient(
					ldaOrFactory.getFactory(), sec);
			storageClient = factClient.createSMS(ldaOrFactory.getStorageType(),
					dSmsId, null);
			catalogue.registerCreatedSms(storageClient.getEPR(),
					ldaOrFactory.getFactory(), dSmsId);
			physicalFilename = ldaOrFactory.getPhysicalFilename();
		} else
		{
			if (storageClient == null)
				storageClient = getStorageClient(ldaOrFactory.getLda(), client, k);
			if (ldaOrFactory != null)
				lda = ldaOrFactory.getLda();
			physicalFilename = lda.getPhysicalName();
		}
		Utils.makeParentDirs(physicalFilename, storageClient);
		return new PhysicalFilenameAndClient(physicalFilename, storageClient);
	}

	/**
	 * Lookup for a file and try to auto-synchronize SMS and dCatalogue
	 * 
	 * @param cat
	 *            - catalogue client
	 * @param path
	 *            - logical path
	 * @param enableSync2
	 *            - if true, syncLookup tries to remove from dCatalogue file
	 *            which doesn't exist in sms
	 * @throws Exception
	 */
	public static LdaWithStorageClient syncLookup(CatalogueClient cat,
			String path, String dSmsId, Client client, boolean sync1,
			boolean sync2, Kernel k) throws Exception
	{
		Lda lda = null;

		if (!sync1)
			lda = cat.lookup(path, dSmsId);
		// sync 1 - file in sms but not in dCat
		else
		{
			try
			{
				lda = cat.lookup(path, dSmsId);
			} catch (FileNotExistsFault e)
			{
				lda = checkIfFileExistsOnSms(cat, path, dSmsId, client, k);
			} catch (InvalidPathFault e)
			{
				lda = checkIfFileExistsOnSms(cat, path, dSmsId, client, k);
			}
		}

		StorageClient storage = getStorageClient(lda, client, k);
		if (sync2)
		{
			// sync 2 - file in dCat but deleted from sms
			GridFileType file = null;
			try
			{
				file = storage.listProperties(lda.getPhysicalName());
			} catch (FileNotFoundException e)
			{
			}
			if (file == null)
			{
				cat.remove(path, dSmsId);
				throw FileNotExistsFault.createFault();
			}
		}
		return new LdaWithStorageClient(lda, storage);
	}

	public static LdaWithStorageClient syncLookup(CatalogueClient cat,
			String path, String dSmsId, Client client, Kernel kernel) throws Exception
	{
		boolean sync1 = isSyncEnabled(kernel, 1);
		boolean sync2 = isSyncEnabled(kernel, 2);
		return syncLookup(cat, path, dSmsId, client, sync1, sync2, kernel);
	}

	/**
	 * Check if file given by path exists on any SMS. If exists, file will be
	 * added to dCatalogue and LDA is returned.
	 * 
	 * @param cat
	 * @param path
	 * @param dSmsId
	 * @param client
	 * @return LDA if file exists, null otherwise.
	 * @throws Exception
	 */
	public static Lda checkIfFileExistsOnSms(CatalogueClient cat, String path,
			String dSmsId, Client client, Kernel kernel) throws Exception
	{
		final int THREADS_LIMIT = 4;

		Stack<Lda> ldas = new Stack<Lda>();
		for (Lda l : cat.lookupDir("/", dSmsId))
			// if(!l.getSms().getAddress().isNil())
			ldas.add(l);

		ExecutorService executor = ThreadingServices.get(kernel).getExecutorService();
		List<Future<Lda>> futures = new ArrayList<Future<Lda>>();
		String xlogin = client.getXlogin().getUserName();
		for (int i = 0; i < THREADS_LIMIT; i++)
			futures.add(executor.submit(new CheckFileOnSmsCallable(ldas, cat,
					path, dSmsId, xlogin, client, kernel)));

		Lda lda = null;
		for (Future<Lda> future : futures)
			try
			{
				if (future.get() != null)
					lda = future.get();
			} catch (Exception e)
			{
				logger.error("Can't connect to SMS", e.getCause());
			}

		if (lda != null)
			return lda;
		else
			throw FileNotExistsFault.createFault();
	}

	/**
	 * Get file URL from LDA and given protocol.
	 * 
	 * @param lda
	 * @param protocol
	 * @return
	 */
	public static String getUrl(Lda lda, String protocol)
	{
		return getUrl(lda.getSms(), lda.getPhysicalName(), protocol);
	}

	/**
	 * Get file URL from SMS EPR, filename and given protocol.
	 * 
	 * @param lda
	 * @param protocol
	 * @return
	 */
	public static String getUrl(EndpointReferenceType smsEpr, String filename,
			String protocol)
	{
		StringBuilder builder = new StringBuilder();
		builder.append(protocol);
		builder.append(':');
		builder.append(smsEpr.getAddress().getStringValue());
		builder.append("#/");
		builder.append(filename);
		return builder.toString();
	}

	/**
	 * Make parent directories for filename on given storage.
	 * 
	 * @param filename
	 * @param storage
	 * @throws BaseFault
	 * @throws FileNotFoundException
	 */
	public static void makeParentDirs(String filename, StorageClient storage)
			throws BaseFault, FileNotFoundException
	{
		String parentDir = filename.substring(0, filename.lastIndexOf('/'));
		try
		{
			GridFileType d = storage.listProperties(parentDir);
			if(d != null && d.getIsDirectory())
				return;
		}
		catch(FileNotFoundException e) { } //ok, we'll create it
		
		String[] pathAr = filename.split("/");
		List<String> pathList = new ArrayList<String>();
		for (String s : pathAr)
			if (!s.isEmpty())
				pathList.add(s);
		String path = "";
		for (int i = 0, length = pathList.size(); i < length - 1; i++)
		{
			path += "/" + pathList.get(i);
			try
			{
				storage.listProperties(path);
			} catch (FileNotFoundException e)
			{
				storage.createDirectory(path);
			}
		}
	}

	/**
	 * Get catalogue client. If it's possible, Catalogue connection will be made
	 * on behalf of user.
	 * 
	 * @param client
	 * @return
	 * @throws Exception
	 */
	public static CatalogueClient getCatalogueClientFromRegistry(Client client, Kernel k) throws Exception
	{
		CatalogueClient cat;
		
		List<EndpointReferenceType> eprs = Utils.listServices(k, Catalogue.CATALOGUE_PORT);
		if (eprs.size() == 0)
			throw new Exception("No dCatalogue in registry");

		for (EndpointReferenceType epr : eprs)
		{
			cat = getCatalogueClient(client, epr, k);
			if(cat != null)
				return cat;
		}
		throw new Exception("Can't connect to any dCatalogue");
	}

	
	public static CatalogueClient getCatalogueClient(Client client, EndpointReferenceType epr, Kernel kernel) throws Exception
	{
		ISecurityProperties sec = null;
		if (client == null)
			sec = kernel.getSecurityProperties().clone();
		else
			try
			{
				sec = kernel.getSecurityProperties().clone();
				String catalogDN=WSUtilities.extractServerIDFromEPR(epr);
				if(catalogDN!=null)
					ETDAssertionForwarding.configureETDChainExtension(client, sec, new X500Principal(catalogDN));
			}
			catch(Exception e)
			{
				logger.warn("Error in getting trust delegation", e);
				sec = kernel.getSecurityProperties();
			}

		sec.setProperty(HttpUtils.CONNECT_TIMEOUT, "600000");
		sec.setProperty(HttpUtils.SO_TIMEOUT, "600000");

		try
		{
			CatalogueClient cat;
			cat = new CatalogueClient(epr, sec);
			cat.getCurrentTime();
			return cat;
		} catch (Exception e)
		{
			logger.error("Can't connect to the dCatalogue", e);
			return null;
		}
	}

	/**
	 * Get storage client. If it's possible, connection will be made on behalf
	 * of user.
	 * 
	 * @param lda
	 * @param client
	 * @return
	 * @throws Exception
	 */
	public static StorageClient getStorageClient(Lda lda, Client client, Kernel kernel)
			throws Exception
	{
		EndpointReferenceType epr = lda.getSms();
		ISecurityProperties sec;
		String dn=WSUtilities.extractServerIDFromEPR(epr);
		try
		{
			sec = kernel.getSecurityProperties().clone();
			if(dn==null){
				ETDAssertionForwarding.configureETD(client, sec);
			}
			else{
				ETDAssertionForwarding.configureETDChainExtension(client, sec, new X500Principal(dn));
			}
		} catch (Exception e)
		{
			logger.warn("Error in getting trust delegation.", e);
			sec = kernel.getSecurityProperties();
		}

		sec.setProperty(HttpUtils.CONNECT_TIMEOUT, "600000");
		sec.setProperty(HttpUtils.SO_TIMEOUT, "600000");
		
		try
		{
			return new StorageClient(epr, sec);
		} catch (Exception e)
		{
			logger.error(
					"Error in connecting to the Storage " + lda.getSmsAddress(),
					e);
			throw e;
		}
	}

	/**
	 * Get services with given port from the registry.
	 * 
	 * @param port
	 * @return
	 * @throws Exception
	 */
	static List<EndpointReferenceType> listServices(Kernel k, QName port)
			throws Exception
	{
		List<EndpointReferenceType> eprs;
		RegistryHandler rh=RegistryHandler.get(k);
		ExternalRegistryClient reg = rh.getExternalRegistryClient();
		if (reg != null)
			eprs = reg.listServices(port);
		else
		{
			eprs = new ArrayList<EndpointReferenceType>();
			for (EntryType e : rh.getRegistryClient()
					.listEntries())
			{
				EndpointReferenceType epr = e.getMemberServiceEPR();
				if (WSServerUtilities.extractInterfaceName(epr).equals(port))
					eprs.add(epr);
			}
		}
		return eprs;
	}

	public static class LdaWithStorageClient
	{
		private Lda lda;
		private StorageClient storage;

		private LdaWithStorageClient(Lda lda, StorageClient storage)
		{
			this.lda = lda;
			this.storage = storage;
		}

		public Lda getLda()
		{
			return lda;
		}

		public StorageClient getStorage()
		{
			return storage;
		}
	}

	public static class PhysicalFilenameAndClient
	{
		private StorageClient client;
		private String physicalFilename;

		public PhysicalFilenameAndClient(String physicalFilename,
				StorageClient client)
		{
			this.client = client;
			this.physicalFilename = physicalFilename;
		}

		public StorageClient getClient()
		{
			return client;
		}

		public String getPhysicalFilename()
		{
			return physicalFilename;
		}
	}

	private static Boolean isSyncEnabled(Kernel k, int type)
	{
		if (type == 1)
			return k.getBooleanProperty(DSMSImpl.ENABLE_SYNC_1);
		else if (type == 2)
			return k.getBooleanProperty(DSMSImpl.ENABLE_SYNC_2);
		else
			return null;
	}
}
