/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
package de.fzj.unicore.uas.impl.registry;

import java.io.InputStream;
import java.util.Calendar;

import javax.servlet.http.HttpServletResponse;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.log4j.Logger;
import org.oasisOpen.docs.wsrf.rp2.GetResourcePropertyDocument;
import org.oasisOpen.docs.wsrf.rp2.GetResourcePropertyResponseDocument;
import org.oasisOpen.docs.wsrf.sg2.AddDocument;
import org.oasisOpen.docs.wsrf.sg2.MemberEPRDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.persist.PersistenceException;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.Home;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.exceptions.ResourceUnknownException;
import de.fzj.unicore.wsrflite.impl.InstanceChecker;
import de.fzj.unicore.wsrflite.utils.WSServerUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.sg.impl.ServiceGroupEntryImpl;
import eu.unicore.security.util.client.HttpUtils;
import eu.unicore.security.util.client.IAuthenticationConfiguration;

/**
 * Checks and refreshes a registry entry in the local registry.
 * 
 * @author demuth
 * @author schuller
 */
public class RegistryEntryUpdater implements InstanceChecker {

	private static final Logger logger=LogUtil.getLogger(LogUtil.SERVICES,RegistryEntryUpdater.class);
	
	public boolean check(Home home, String id)throws ResourceUnknownException,PersistenceException{
		Calendar c=home.getTerminationTime(id);
		if(logger.isDebugEnabled()){
			logger.debug("Checking <"+home.getServiceName()+">"+id+" TT = "+(c!=null?c.getTime():"none"));
		}
		//if for some reason the TT is null, force a refresh (in contrast to the usual expiry check)
		return c==null? true : (c.compareTo(Calendar.getInstance())<=0);
	}


	private final byte[] buf=new byte[1024];

	/**
	 * For the current registry entry, it is checked whether the 
	 * corresponding member service (i.e. the service registered in the local registry)
	 * is still alive. 
	 * 
	 * <ul>
	 *  <li>If the member service is no longer alive, its entry is refreshed, i.e. re-added 
	 * to the local registry.
	 * 
	 *  <li>If the member service is alive, but can't be reached (say, if the gateway is down),
	 * the entry is not refreshed, but stays alive and will be re-checked. 
	 * 
	 *  <li>If the member service is gone, the registry entry is removed.
	 *</ul>
	 *
	 * @return <code>true</code> if the registry entry is still valid, <code>false</code> 
	 * if the service is gone and the registry entry was removed 
	 */
	public boolean process(Home home, String id) {
		if(home.isShuttingDown())return true;
		Kernel kernel=home.getKernel();
		String serviceName = home.getServiceName();
		try{
			RegistryEntryImpl entry=(RegistryEntryImpl)home.get(id);

			GetResourcePropertyDocument req=GetResourcePropertyDocument.Factory.newInstance();
			req.setGetResourceProperty(ServiceGroupEntryImpl.RPMemberEPR);
			GetResourcePropertyResponseDocument res=entry.GetResourceProperty(req);
			MemberEPRDocument memberEprDoc = MemberEPRDocument.Factory.parse(res.getGetResourcePropertyResponse().newInputStream());
			EndpointReferenceType memberEpr = memberEprDoc.getMemberEPR(); 
			String memberAddress=memberEpr.getAddress().getStringValue();
			
			//check that URL is still basically correct (e.g. hostname, port) 
			if(!checkBasicCorrectness(memberAddress, kernel)){
				logger.debug("Member address "+memberAddress+" is no longer valid, destroying registry entry.");
				entry.destroy();
				home.destroyWSRFServiceInstance(id);
				//instance is invalid and should be removed from all checks
				return false;
			}
			
			// for WSRF service instances: check if instance is still available in this container
			if(memberAddress.toLowerCase().contains("?res="))
			{
				try 
				{
					if(logger.isDebugEnabled()){
						logger.debug("Alive check: "+memberAddress);
					}
					Home serviceHome = kernel.getHome(WSServerUtilities.extractServiceName(memberEpr));
					serviceHome.get(WSServerUtilities.extractResourceID(memberEpr));

				} 
				catch (Exception e) 
				{
					LogUtil.logException("Could not find WSRF instance in virtual machine. Destroying entry.",e,logger);
					entry.destroy();
					home.destroyWSRFServiceInstance(id);
					//instance is invalid and should be removed from all checks
					return false;
				}
			}
			
			//make an external call to make sure the service is reachable (incl. gateway)
			try {
				String uri=WSServerUtilities.makeAddress(serviceName,kernel);
				IAuthenticationConfiguration auth=kernel.getSecurityProperties();
				HttpClient client=HttpUtils.createClient(uri,auth, kernel.getProperties());
				
				GetMethod get=new GetMethod(uri+"?wsdl");
				try{
					client.executeMethod(get);
					if(get.getStatusCode()!=HttpServletResponse.SC_OK){
						logger.warn("Error reaching service, not publishing entry to external registry. Error was: "+get.getStatusLine().toString());
						return true;
					}
					//read (and discard) the response to avoid exceptions on the gateway
					InputStream is=get.getResponseBodyAsStream();
					while(is.read(buf)!=-1);
				}
				finally{
					get.releaseConnection();
				}
			} catch (Exception e) {
				logger.warn("Could not reach service via web service call. Not publishing entry to external registry.",e);
				//instance is still valid
				return true;
			}

			try
			{
				logger.debug("Trying to re-add service to registry: " +memberAddress);						
				reAdd(home.getKernel(),memberEpr);
				logger.info("Refreshed registry entry for: " +memberAddress);
			}
			catch(Exception e)
			{						
				LogUtil.logException("Error re-adding service entry: ",e,logger);
			}		
		}

		catch(Exception e){
			LogUtil.logException("Could not update WSRF instance of type "+serviceName+" and id: "+id,e,logger);
		}

		// instance is still valid
		return true;
	}

	/**
	 * 
	 * @param url - the member service URL to check
	 * @return <code>true</code> if URL looks OK
	 */
	protected boolean checkBasicCorrectness(String url, Kernel k){
		String baseURL=k.getBaseURL();
		return url!=null && url.startsWith(baseURL);
	}

	protected void reAdd(Kernel kernel, EndpointReferenceType memberEpr) throws Exception
	{
		AddDocument add = AddDocument.Factory.newInstance();
		add.addNewAdd().setMemberEPR(memberEpr);
		add.getAdd().setInitialTerminationTime(null);
		RegistryHandler.get(kernel).getRegistryClient().addRegistryEntry(add);
	}


}
