/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/

package de.fzj.unicore.uas.impl.registry;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;

import org.apache.log4j.Logger;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.persist.impl.LockSupport;
import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.client.LocalRegistryClient;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.Home;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.events.PropertyFileChangeEvent;
import de.fzj.unicore.wsrflite.exceptions.ResourceNotCreatedException;
import de.fzj.unicore.wsrflite.exceptions.ResourceUnknownException;
import de.fzj.unicore.wsrflite.security.UASSecurityProperties;
import de.fzj.unicore.wsrflite.utils.WSServerUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.client.ExternalRegistryClient;
import de.fzj.unicore.wsrflite.xmlbeans.client.MultiWSRFClient;
import de.fzj.unicore.wsrflite.xmlbeans.client.RegistryClient;
import de.fzj.unicore.wsrflite.xmlbeans.impl.WSResourceImpl;

/**
 * Manages registry info. It is used to obtain a client for both
 * internal and external registries
 *
 * @author demuth
 * @author schuller
 */
public class RegistryHandler {

	protected static int initial = 20, period = 20;

	/**
	 * the set of external registry URLs. It is updated whenever the property file
	 * changes using the {@link #updateExternalRegistryURLs()} method
	 */
	private final Set<String> externalRegistryURLs = new HashSet<String>();

	private LocalRegistryClient registryClient = null;

	private Boolean isGlobalRegistry = null; 

	protected ScheduledExecutorService scheduledExecutor = null;

	private static final Logger logger=LogUtil.getLogger(LogUtil.SERVICES,RegistryHandler.class);

	private final Kernel kernel;
	
	private RegistryHandler(Kernel kernel)throws Exception{
		this.kernel=kernel;
		initRegistry();
	}
	
	private void initRegistry()throws Exception{

		Observer obs=new Observer(){
			public void update(Observable obs, Object o){
				if(o instanceof PropertyFileChangeEvent){
					logger.info("Properties file change: updating external registry settings.");
					synchronized(this){
						externalRegistryURLs.clear();
						updateExternalRegistryURLs();
					}
				}
			}
		};

		kernel.addObserver(obs);

		//	set up default registry...
		String resID = "default_registry";

		updateExternalRegistryURLs();
		Home regHome=kernel.getHome(UAS.REG);
		if(regHome==null){
			logger.info("No Registry service configured for this site.");
			return;
		}
		//get "global" lock
		LockSupport ls=kernel.getPersistenceManager().getLockSupport();
		Lock regLock=ls.getOrCreateLock(RegistryHandler.class.getName());
		if(regLock.tryLock()){
			try{
				//check if default registry already exists
				try{
					regHome.get(resID);
					logger.debug("Registry has already been set up.");
				}catch(ResourceUnknownException e){
					try {
						logger.debug("Trying to start up registry.");
						createRegistryInstance(regHome);
					} catch (Exception ex) {
						logger.warn("Could not start up registry!",ex);
					}
				}
			}
			finally{
				regLock.unlock();
			}
		}

		//backwards compatibility: switch off registry access control if not already 
		//configured otherwise
		if(kernel.getProperty(UASSecurityProperties.UAS_CHECKACCESS+"."+UAS.REG)==null){
			kernel.setProperty(UASSecurityProperties.UAS_CHECKACCESS+"."+UAS.REG,"false");
			logger.info("Backwards compatibility: Switching off access control on internal registry."+
					"Set property '"+UASSecurityProperties.UAS_CHECKACCESS+"."+UAS.REG
					+"' to true to enable access control.");
		}

		try{
			registryClient = new LocalRegistryClient("default_registry",kernel);	
		}catch(Exception ex){
			logger.fatal("Could not create client for local registry!",ex);
		}

	}

	private static void createRegistryInstance(Home regHome)throws ResourceNotCreatedException{
		String resID = "default_registry";
		Map<String,Object>map=new HashMap<String,Object>();
		map.put(WSResourceImpl.INIT_UNIQUE_ID,resID);
		//set lifetime to 36 months
		Calendar c1=new GregorianCalendar();
		c1.add(Calendar.MONTH,36);
		map.put(WSResourceImpl.INIT_INITIAL_TERMINATION_TIME,c1);
		regHome.createWSRFServiceInstance(map);
		logger.debug("Added '"+resID+"' resource to service "+UAS.REG);
	}

	/**
	 * this is only used internally and by the JMX code
	 * Application code should retrieve a {@link RegistryClient} using {@link #getExternalRegistryClient()}
	 * @returns the current list of external registry URLs
	 */
	public synchronized String[] getExternalRegistryURLs(){
		return externalRegistryURLs.toArray(new String[externalRegistryURLs.size()]);
	}

	/**
	 * updates the set of external registry URLs (without clearing it first)
	 */
	private void updateExternalRegistryURLs(){
		if(isGlobalRegistry()) logger.debug("This is a global registry. No external registry used.");
		if(!usesExternalRegistry()) logger.debug("Usage of external registry is switched off by property "+UAS.EXTERNAL_REGISTRY_USE +". No external registry used.");

		if(!isGlobalRegistry() && usesExternalRegistry()) {
			logger.debug("Determining external registry address(es) ...");

			if(Boolean.parseBoolean(UAS.REGISTRY_AUTODISCOVER)) {
				logger.debug("Setting up auto-discovery of external registry.");
				if(scheduledExecutor == null) {
					logger.debug("Scheduling repeated auto-discovery.");
					RegistryFinder finder = new RegistryFinder(kernel);
					finder.run();
					scheduledExecutor = Executors.newScheduledThreadPool(1);
					scheduledExecutor.scheduleAtFixedRate(finder,period,initial,TimeUnit.SECONDS);
				}
			}
			else logger.debug("Auto-discovery of external registry switched off.");

			synchronized (externalRegistryURLs) {
				// if autodiscovery switched off or failed apply user given registry address as fallback
				if(externalRegistryURLs.size() == 0) {
					logger.debug("Trying to use manual settings for external registry.");
					for(String registryURL: getConfiguredRegistryURLs()){
						if(registryURL!=null && registryURL.length()>0){
							String u=registryURL.trim();
							externalRegistryURLs.add(u);
							logger.info("Using registry: "+u);
						}
					}
				}
			}
			if (externalRegistryURLs.size()==0){
				logger.warn("No external registry URLs are defined!");
			}
		}
	}

	private String[] getConfiguredRegistryURLs(){
		//registries configured by URL
		List<String>keys=getPropertyKeys(UAS.EXTERNAL_REGISTRY_KEY);
		String[] res=new String[keys.size()];
		for(int i=0;i<res.length;i++){
			res[i]=kernel.getProperty(keys.get(i));
		}
		return res;
	}

	private List<String> getPropertyKeys(String pattern){
		List<String>res=new ArrayList<String>();
		for(Object o: kernel.getProperties().keySet()){
			if(o.toString().startsWith(pattern))res.add((String)o);
		}
		return res;
	}

	private static EndpointReferenceType makeEPR(String url){
		String wsaTo = url;
		String resID = url.split("=")[1];
		EndpointReferenceType epr=EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(wsaTo);
		WSServerUtilities.addUGSRefparamToEpr(epr,resID);
		return epr;
	}

	/**
	 * set the external registry URL
	 * @return
	 */
	protected synchronized void addExternalRegistryURL(String url){
		externalRegistryURLs.add(url);
	}

	/**
	 * is this service container hosting the global registry?
	 */
	public synchronized boolean isGlobalRegistry(){
		if(isGlobalRegistry==null){
			Home regHome=kernel.getHome(UAS.REG);
			if(regHome!=null && regHome instanceof RegistryHomeImpl){
				isGlobalRegistry = true;
			}
			else{
				isGlobalRegistry = false;
			}
		}
		return isGlobalRegistry;
	}

	/**
	 * is this service container contacting a global registry?
	 * @return <code>true</code> if property {@link UAS#EXTERNAL_REGISTRY_USE} is true
	 */
	public boolean usesExternalRegistry(){
		return Boolean.parseBoolean(kernel.getProperty(UAS.EXTERNAL_REGISTRY_USE));
	}

	/**
	 * get a client for the internal registry 
	 * @return
	 */
	public LocalRegistryClient getRegistryClient()throws Exception{
		return registryClient;
	}

	/**
	 * get a client for talking to the external registries
	 * @see MultiWSRFClient 
	 * @return {@link ExternalRegistryClient}
	 */ 
	public ExternalRegistryClient getExternalRegistryClient()throws Exception{
		if(!usesExternalRegistry())return null;
		ExternalRegistryClient reg=new ExternalRegistryClient();
		synchronized(externalRegistryURLs){
			for(String url: externalRegistryURLs){
				try{
					reg.addClient(new RegistryClient(url,makeEPR(url),kernel.getSecurityProperties()));
				}catch(Exception e){
					logger.error("Could not create client for external registry at <"+url+">",e);
				}
			}
		}
		reg.setMode(ExternalRegistryClient.MULTICAST);
		return reg;
	}


	public static synchronized RegistryHandler get(Kernel k)throws Exception{
		RegistryHandler h=k.getAttribute(RegistryHandler.class);
		if(h==null){
			h = new RegistryHandler(k);
			k.setAttribute(RegistryHandler.class, h);
		}
		return h;
	}


}
