package de.fzj.unicore.uas.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.impl.sms.StorageManagementHomeImpl.StorageTypes;
import de.fzj.unicore.wsrflite.utils.PropertyHelper;

/**
 * parses the supplied properties, these may contain
 * storage descriptions as follows:<br/>
 * <code>
 * uas.targetsystem.storage.N=XXXX
 * uas.targetsystem.storage.N.type=...
 * uas.targetsystem.storage.N.path=...
 * </code>
 * where <code>N</code> is a string for disambiguating various storages 
 * (e.g. "1", "2", etc may be used)
 * <br/>
 * 
 * Examples:
 * 
 *  <code>
 * uas.targetsystem.storage.1=DEISAHOME
 * uas.targetsystem.storage.type.1=VARIABLE
 * uas.targetsystem.storage.path.1=DEISA_HOME
 * </code>
 * will attach a storage defined per-user by evaluating the 
 * $DEISA_HOME environment variable <br/>
 * 
 *  <code>
 * uas.targetsystem.storage.2=TEMP
 * uas.targetsystem.storage.type.2=FIXEDPATH
 * uas.targetsystem.storage.path.2=/tmp
 * </code>
 * will attach a storage defined using the supplied path<br/>
 * 
 * TODO handle detailed permissions
 */
public class AddOnStorageDescriptionFactory{
	
	public static final String PREFIX="uas.targetsystem.storage.";
	public static final String SUFFIX_TYPE=".type";
	public static final String SUFFIX_PATH=".path";
	public static final String SUFFIX_CLASS=".class";
	public static final String SUFFIX_PROTOCOLS=".protocols";
	public static final String SUFFIX_FILTER_LISTING="filterFiles";
	
	List<AddOnStorageDescription>list=new ArrayList<AddOnStorageDescription>();
	
	/**
	 * Will parse UAS.getProperties()
	 * @deprecated
	 */
	public AddOnStorageDescriptionFactory(){
		this(UAS.getProperties());
	}
	/**
	 * @param props - the properties to parse (usually UAS.getProperties())
	 */
	public AddOnStorageDescriptionFactory(Properties props){
		for(Object k: props.keySet()){
			String key=((String)k).trim();
			if (!key.startsWith(PREFIX) || key.length() < PREFIX.length()
					|| key.equals(UAS.TSS_FORCE_UNIQUE_STORAGE_IDS))
				continue;
			
			String id = key.substring(PREFIX.length());
			if (id.contains("."))
				continue;
			
			//we have storage id in 'id'. 
			String name=props.getProperty(key);
			String type=props.getProperty(PREFIX+id+SUFFIX_TYPE);
			String path=props.getProperty(PREFIX+id+SUFFIX_PATH);
			String clazz=props.getProperty(PREFIX+id+SUFFIX_CLASS);
			String protocols_property=PREFIX+id+SUFFIX_PROTOCOLS;

			PropertyHelper helper = new PropertyHelper(props, PREFIX+id+".");
			Map<String, String> storageProperties = createExtraSettings(
					helper.getFilteredMap(), id);
			
			try {
				AddOnStorageDescription asd=new AddOnStorageDescription(name,path,type,
						clazz,protocols_property,storageProperties);
				list.add(asd);
			} catch(Exception ex) {
				LogUtil.getLogger(LogUtil.SERVICES, AddOnStorageDescriptionFactory.class).warn(
						"Invalid storage description : "+key);
			}
		}
	}

	/**
	 * Remove storage id from the property keys to form a uniform property names.
	 * @param original
	 * @return
	 */
	private static Map<String, String> createExtraSettings(Map<String, String> original, 
			String storageId) {
		Map<String, String> ret = new HashMap<String, String>();
		int len = (PREFIX + storageId + ".").length();
		for (String key: original.keySet()) {
			String newKey = PREFIX + key.substring(len);
			ret.put(newKey, original.get(key));
		}
		return ret;
	}
	
	public List<AddOnStorageDescription> getDescriptions() {
		return list;
	}
	
	public class AddOnStorageDescription{
		
		private final String name, pathSpec, clazz, protocols_property;
		private StorageTypes type;
		private Map<String, String> additionalProperties;
		
		public AddOnStorageDescription(String name, String pathSpec, String type, String clazz, 
				String protocols_property, Map<String, String> additionalProperties){
			this.name=name;
			this.pathSpec=pathSpec;
			this.clazz=clazz;
			this.protocols_property=protocols_property;
			this.type=StorageTypes.valueOf(type);
			this.additionalProperties = additionalProperties;
			
			if (type.equals(StorageTypes.CUSTOM) && clazz == null)
				throw new IllegalArgumentException("No class defined for the CUSTOM storage");
		}

		public String getName() {
			return name;
		}
		
		public String getProtocolsProperty() {
			return protocols_property;
		}
		
		public String getPathSpec() {
			return pathSpec;
		}

		public String getType() {
			return type.toString();
		}
		
		public String getStorageClass() {
			return clazz;
		}
		
		public StorageTypes getStorageType() {
			return type;
		}
		
		public String toString(){
			return "Storage description: name="+name+", type="+type+", pathSpec="+pathSpec+", protocols="+protocols_property;
		}

		public Map<String, String> getAdditionalProperties() {
			return additionalProperties;
		}
	}
	
}

