package de.fzj.unicore.uas.dsms;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.unigrids.services.atomic.types.GridFileType;

import pl.edu.icm.dsms.catalogue.db.InitDB;
import de.fzj.unicore.uas.Base;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.TransferControllerClient;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;

public class TestDsmsBasics extends DsmsTestBase
{
	protected StorageClient dsms, sms;

	private String dataMap="src/test/resources/dcatalogue/datamap.properties";
	
	@BeforeClass
	public static void startKernel()
	{
		try
		{
			Base.initDirectories();
			System.out.println("Starting kernel");
			AbstractTestBase.startKernel();
		} catch (Exception e)
		{
			e.printStackTrace();
			fail("Failed to start kernel");
		}
	}

	@AfterClass
	public static void shutdownKernel()
	{
		try
		{
			System.out.println("Shutting down kernel");
			AbstractTestBase.shutdownKernel();
		} catch (Exception e)
		{
			e.printStackTrace();
			fail("Failed to shutdown kernel");
		}
	}

	@Before
	public void setUp() throws Exception
	{
		cleanStorageDirs();
		InitDB.initDB(dataMap);
		createSms();
		createCatalogue();
		createDsms();

		dsms = getDsmsClient();
		sms = getSmsClient();
	}

	@After
	public void tearDown()
	{
		try
		{
			sms = null;
			dsms = null;

			destroyDsms();
			destroyCatalogue();
			destroySms();
		} catch (Exception e)
		{
			e.printStackTrace();
			fail("Unable to tear down");
		}
	}

	@Test
	public void testCreateDir() throws BaseFault, FileNotFoundException
	{
		try{
			dsms.listProperties("dir");
			//exists already, so clean it up
			dsms.delete("dir");
		}catch(FileNotFoundException e){ /* ok */}
		//create dir
		dsms.createDirectory("dir");
		assertTrue(dsms.listProperties("dir").isSetIsDirectory());
	}

	@Test
	public void testImportExport() throws IOException, Exception
	{	
		//import
		byte body[] = "test file body".getBytes();
		dsms.getImport("dir/test").write(body);

		//export
		byte recv[] = dsms.getExport("dir/test").read(0, 1, body.length, 0);
		assertTrue(new String(body).equals(new String(recv)));
	}

	@Test
	public void testList() throws IOException, Exception
	{
		dsms.getImport("dir/test2").write("test 2".getBytes());
		dsms.getImport("dir/test3").write("test 3".getBytes());
		dsms.getImport("dir/test4").write("test 3".getBytes());
		dsms.createDirectory("dir/subdir");
		GridFileType[] list = dsms.listDirectory("dir");
		Assert.assertTrue(Arrays.toString(list), compareFileList(list, new String[] {"test2", "test3", "test4", "subdir"}));
	}

	@Test
	public void testListProperties() throws IOException, Exception
	{
		dsms.createDirectory("dir");
		byte body[] = "test file body".getBytes();
		dsms.getImport("dir/test").write(body);

		//list properties
		GridFileType fileType = dsms.listProperties("dir/test");
		assertTrue(body.length == fileType.getSize());
		assertTrue(dsms.listProperties("dir").getIsDirectory());

		//set permission
		dsms.changePermissions("dir/test", true, true, true);
		//assertTrue(!dsms.listProperties("dir/test").getPermissions().getWritable());
		dsms.changePermissions("dir/test", true, true, false);
		//assertTrue(dsms.listProperties("dir/test").getPermissions().getWritable());
	}

	@Test
	public void testRename() throws IOException, Exception
	{
		dsms.createDirectory("dir");
		dsms.getImport("dir/test").write("test".getBytes());

		//rename
		dsms.rename("dir/test", "dir/new_test");
		Boolean exceptionOccured = false;
		try { dsms.listProperties("dir/test"); }
		catch(FileNotFoundException e) { exceptionOccured = true; }
		assertTrue(exceptionOccured);
		dsms.listProperties("dir/new_test");
	}

	@Test
	public void testFind() throws IOException, Exception
	{
		//find
		dsms.getImport("new_test").write("test".getBytes());
		GridFileType[] list = dsms.find("/", true, "new_test", false, null, null);
		assertTrue(compareFileList(list, new String[] {"new_test"}));
	}

	@Test
	public void testSend() throws IOException, Exception
	{
		//sendFile && recvFile
		String foreignUrl = "BFT:" + sms.getEPR().getAddress().getStringValue() + "#" + "/sentTest";
		byte[] body = "test file body".getBytes();
		dsms.getImport("fileToSend").write(body);

		TransferControllerClient tcc = dsms.sendFile("fileToSend", foreignUrl);
		while(!tcc.isComplete() && !tcc.hasFailed())
			Thread.sleep(1000);
		assertTrue(tcc.isComplete());
		assertTrue(!tcc.hasFailed());
		byte[] recv = sms.getExport("/sentTest").read(0, 1, body.length, 0);
		assertTrue(new String(body).equals(new String(recv)));
	}

	@Test
	public void testRecv() throws IOException, Exception
	{
		String foreignUrl = "BFT:" + sms.getEPR().getAddress().getStringValue() + "#" + "/fileToRecv";
		byte[] body = "test file body".getBytes();
		sms.getImport("/fileToRecv").write(body);

		TransferControllerClient tcc = dsms.fetchFile(foreignUrl, "/receivedFile");
		while(!tcc.isComplete() && !tcc.hasFailed())
			Thread.sleep(1000);
		assertTrue(tcc.isComplete());
		assertTrue(!tcc.hasFailed());
		byte[] recv = dsms.getExport("/receivedFile").read(0, 1, body.length, 0);
		assertTrue(new String(body).equals(new String(recv)));
	}

	@Test
	public void testCopy() throws IOException, Exception
	{
		byte[] body = "test file body".getBytes();
		dsms.getImport("test").write(body);
		dsms.copy("/test", "/copiedFile");
		byte[] recv = dsms.getExport("/copiedFile").read(0, 1, body.length, 0);
		assertTrue(new String(body).equals(new String(recv)));
	}

	@Test
	public void testFreeSpace() throws Exception
	{
		Long freeSpace = dsms.getAvailableSpace();
		assertTrue(freeSpace != null);
		System.out.println("Free space: "+freeSpace);
		assertTrue(freeSpace > 0);
	}
}
