package de.fzj.unicore.uas.impl.registry;

import java.util.Calendar;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.xmlbeans.XmlDateTime;
import org.oasisOpen.docs.wsrf.sg2.AddDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.Registry;
import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.util.RegistryStartup;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.impl.DefaultHome;
import de.fzj.unicore.wsrflite.utils.WSServerUtilities;
import de.fzj.unicore.wsrflite.xfire.JettyTestCase;
import de.fzj.unicore.wsrflite.xfire.XFireServiceFactory;
import de.fzj.unicore.wsrflite.xmlbeans.client.ExternalRegistryClient;
import de.fzj.unicore.wsrflite.xmlbeans.client.RegistryClient;
import de.fzj.unicore.wsrflite.xmlbeans.sg.ServiceGroupEntry;
import eu.unicore.bugsreporter.annotation.FunctionalTest;

/**
 * 
 * Tests functionality of the shared registry<br/>
 * <ul>
 *  <li>register a service in the registry</li>
 *  <li>re-register a service in the registry</li>
 *  <li>removal of a service when it is no longer available</li>
 *  <li>re-registration with modified content</li>
 * </ul>
 * 
 */
public class TestSharedRegistry extends JettyTestCase{

	@Override
	protected void addServices() throws Exception {
		kernel.setProperty("persistence.directory", "target/data-"+System.currentTimeMillis());
		XFireServiceFactory.createAndDeployService(kernel,UAS.REG,Registry.class,RegistryHomeImpl.class,true);
		XFireServiceFactory.createAndDeployService(kernel,"ServiceGroupEntry",ServiceGroupEntry.class,RegistryEntryHomeImpl.class,true);
	}

	@FunctionalTest(id="testRegistryRegistrationAndCleanup",
					description="Test registration of a service in the registry, and " +
							    "its removal when no longer available")
	public void testRegistryRegistrationAndCleanup()throws Exception{
		
		new RegistryStartup(kernel).run();

		ExternalRegistryClient client=new ExternalRegistryClient();
		EndpointReferenceType epr=EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(kernel.getProperty(Kernel.WSRF_BASEURL)
				+"/Registry?res=default_registry");
		RegistryClient regClient=new RegistryClient(epr,kernel.getSecurityProperties());
		regClient.setUpdateInterval(-1);
		client.addClient(regClient);
		assertTrue(regClient.checkConnection());

		AddDocument in = AddDocument.Factory.newInstance();
		EndpointReferenceType sEpr=in.addNewAdd().addNewMemberEPR();
		sEpr.addNewAddress().setStringValue("http://test");
		QName portType=new QName("foo","bar");
		WSServerUtilities.addPortType(sEpr, portType);
		WSServerUtilities.addServerIdentity(sEpr, "CN=server");
		client.addRegistryEntry(in);

		List<EndpointReferenceType> entries=client.listServices(portType);
		assertNotNull(entries);
		assertEquals(1, entries.size());
		EndpointReferenceType service=entries.get(0);
		String dn=WSServerUtilities.extractServerIDFromEPR(service);
		assertEquals("CN=server",dn);

		//re-register, this time setting TT to small value
		Calendar cal=Calendar.getInstance();
		cal.add(Calendar.SECOND, 1);
		XmlDateTime xcal=XmlDateTime.Factory.newInstance();
		xcal.setCalendarValue(cal);
		in.getAdd().setInitialTerminationTime(cal);

		client.addRegistryEntry(in);
		entries=client.listServices(portType);
		assertNotNull(entries);
		assertEquals(1, entries.size());

		Thread.sleep(3000); 

		//force expiry check and test that the entry is gone
		DefaultHome home=(DefaultHome)kernel.getHome("ServiceGroupEntry");
		home.runExpiryCheckNow();
		regClient.getCurrentTime();
		entries=client.listServices(portType);
		assertNotNull(entries);
		assertEquals(0, entries.size());

	}

	
	@FunctionalTest(id="testUpdateRegistryContent",
			description="Tests that the service description in the registry " +
					    "is properly updated")
	public void testUpdateRegistryContent()throws Exception{
		new RegistryStartup(kernel).run();

		ExternalRegistryClient client=new ExternalRegistryClient();
		EndpointReferenceType epr=EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(kernel.getProperty(Kernel.WSRF_BASEURL)
				+"/Registry?res=default_registry");
		RegistryClient regClient=new RegistryClient(epr,kernel.getSecurityProperties());
		regClient.setUpdateInterval(-1);
		client.addClient(regClient);
		assertTrue(regClient.checkConnection());

		AddDocument in = AddDocument.Factory.newInstance();
		EndpointReferenceType sEpr=in.addNewAdd().addNewMemberEPR();
		sEpr.addNewAddress().setStringValue("http://test");
		QName portType=new QName("foo","bar");
		WSServerUtilities.addPortType(sEpr, portType);
		WSServerUtilities.addServerIdentity(sEpr, "CN=server");
		client.addRegistryEntry(in);

		List<EndpointReferenceType> entries=client.listServices(portType);
		assertNotNull(entries);
		assertEquals(1, entries.size());
		EndpointReferenceType service=entries.get(0);
		String dn=WSServerUtilities.extractServerIDFromEPR(service);
		assertEquals("CN=server",dn);

		//re-register with modified content
		in = AddDocument.Factory.newInstance();
		sEpr=in.addNewAdd().addNewMemberEPR();
		sEpr.addNewAddress().setStringValue("http://test");
		WSServerUtilities.addPortType(sEpr, portType);
		WSServerUtilities.addServerIdentity(sEpr, "CN=new server");
		client.addRegistryEntry(in);

		entries=client.listServices(portType);
		assertNotNull(entries);
		assertEquals(1, entries.size());
		service=entries.get(0);
		dn=WSServerUtilities.extractServerIDFromEPR(service);
		assertEquals("CN=new server",dn);

	}

	@FunctionalTest(id="testReRegistration",
			description="Tests that services can register multiple times, but only a single entry " +
					    "per service is created")
	public void testReRegistration()throws Exception{
		new RegistryStartup(kernel).run();

		ExternalRegistryClient client=new ExternalRegistryClient();
		EndpointReferenceType epr=EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(kernel.getProperty(Kernel.WSRF_BASEURL)
				+"/Registry?res=default_registry");
		RegistryClient regClient=new RegistryClient(epr,kernel.getSecurityProperties());
		regClient.setUpdateInterval(-1);
		client.addClient(regClient);
		assertTrue(regClient.checkConnection());

		AddDocument in = AddDocument.Factory.newInstance();
		EndpointReferenceType sEpr=in.addNewAdd().addNewMemberEPR();
		sEpr.addNewAddress().setStringValue("http://test");
		QName portType=new QName("foo","bar2");
		WSServerUtilities.addPortType(sEpr, portType);
		WSServerUtilities.addServerIdentity(sEpr, "CN=server");
		int size=client.listServices(portType).size();
		for(int i=0; i<51;i++){
			client.addRegistryEntry(in);
			Thread.sleep(100);
			assertEquals("Too many registry entries", size+1, client.listEntries().size());
		}
	}

	public void testManyEntries()throws Exception{
		new RegistryStartup(kernel).run();

		ExternalRegistryClient client=new ExternalRegistryClient();
		EndpointReferenceType epr=EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(kernel.getProperty(Kernel.WSRF_BASEURL)
				+"/Registry?res=default_registry");
		RegistryClient regClient=new RegistryClient(epr,kernel.getSecurityProperties());
		regClient.setUpdateInterval(-1);
		client.addClient(regClient);
		assertTrue(regClient.checkConnection());
		QName portType=new QName("foo","bar3");
		int N=100;
		System.out.println("Adding "+N+" entries into registry.");
		for(int i=0; i<N; i++){
			AddDocument in = AddDocument.Factory.newInstance();
			EndpointReferenceType sEpr=in.addNewAdd().addNewMemberEPR();
			sEpr.addNewAddress().setStringValue("http://test_"+i);
			WSServerUtilities.addPortType(sEpr, portType);
			WSServerUtilities.addServerIdentity(sEpr, "CN=server");
			client.addRegistryEntry(in);
		}
		System.out.println("Checking lookup time for "+N+" entries.");
		long start=System.currentTimeMillis();
		System.out.println("Entries: "+regClient.getResourcePropertiesDocument().getServiceGroupRP().getEntryArray().length);

		int l=client.listServices(portType).size();
		System.out.println("Have "+l+" entries in the registry");
		assertTrue(l>=N);
		long finish=System.currentTimeMillis();
		long took=finish-start;
		System.out.println("Lookup on "+N+" registry entries took : "+(took)+ " ms.");
		assertTrue("Lookup seems to be very slow",took<20*N);
	}

	
}
