/*
 *  UNICORE UDT "receive" application
 *
 *  usage: recvfile server_ip local_filename mode [comm_out comm_in]
 *  
 *  Operation:
 *
 *  1. The application will start a UDT connection to the given server ip, and 
 *  write the local port to the file named comm_out. If not given, the port
 *  is written to stderr.  
 *
 *  2. The application will read the remote port from comm_in (or stdin) 
 *
 *  3. When the connection is thus established, the local file will be received
 *
 *
 *  Authors: Thomas Oistrez, Bernd Schuller (JSC, Forschungszentrum Juelich GmbH)
*/

#ifndef WIN32
   #include <arpa/inet.h>
#else
   #include <winsock2.h>
   #include <ws2tcpip.h>
#endif
#include <fstream>
#include <iostream>
#include <cstdlib>
#include <cstring>
#include <udt.h>
#include <sys/stat.h>
#include <pthread.h>

using namespace std;

/* 
* helper to check if file exists
*/
bool file_exists(char* filename) {
  struct stat fileInfo;
  bool res;
  int stats;

  // try to get the file attributes
  stats = stat(filename,&fileInfo);
  if(stats == 0) {
    res = true;
  } else {
    res = false;
  }
  
  return(res);
}

int main(int argc, char* argv[])
{

  // index of arguments in argv[]
#define ARG_IP 1
#define ARG_LOCALFILE 2
#define ARG_MODE 3
#define ARG_COMM_OUT 4
#define ARG_COMM_IN 5

   if (argc < 4)
   {
      cerr << "usage: recvfile server_ip local_filename mode [comm_out comm_in]" << endl;
      return 1;
   }

   UDTSOCKET fhandle = UDT::socket(AF_INET, SOCK_STREAM, 0);
   sockaddr_in serv_addr;
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_port = 0;

   ofstream myfile;
   bool haveCommFile=false; 
   if(argc>4)haveCommFile=true;

 
 
   if(haveCommFile)
   {
          myfile.open (argv[ARG_COMM_OUT]);
          if(!myfile.is_open())
          {
             cerr << "ERROR: Can't open comm-out file for writing." << endl;
             return 1;
          }
   }

   ostream& out = haveCommFile? myfile : cerr ;

#ifndef WIN32
   if (inet_pton(AF_INET, argv[ARG_IP], &serv_addr.sin_addr) <= 0)
#else
   if (INADDR_NONE == (serv_addr.sin_addr.s_addr = inet_addr(argv[ARG_IP])))
#endif
   {
      out << "ERROR: incorrect network address.\n";
      return 1;
   }
   memset(&(serv_addr.sin_zero), '\0', 8);

   //bind before connect fixes the port
   sockaddr_in client_addr;
   int client_addr_size = sizeof(client_addr);
   client_addr.sin_family = AF_INET;
   client_addr.sin_port = 0;
   client_addr.sin_addr.s_addr = INADDR_ANY;

   if (UDT::ERROR == UDT::bind(fhandle,(sockaddr*)&client_addr, sizeof(client_addr)))
   {
      out << "ERROR: bind: " << UDT::getlasterror().getErrorMessage() << endl;
      out.flush();
      return 1;
   }else{
      if(UDT::ERROR == UDT::getsockname(fhandle,(sockaddr*)&client_addr,&client_addr_size))
      {
         out << "ERROR: getsockname: " << UDT::getlasterror().getErrorMessage() << endl;
         return 1; 
      }

      //write our port
      out << "OUT: " << ntohs(client_addr.sin_port) << endl;

   }

   //get server port
   int srv_port = 0;

   //read peer port from file or stdin
   if(haveCommFile)
   {

          while(!file_exists(argv[ARG_COMM_IN])){
             out << "INFO: Comm in-file not found, waiting..." << endl;             
             sleep(1);
          }

          ifstream infile;
          infile.open(argv[ARG_COMM_IN]);
  	  if(infile.is_open())
          {
             infile >> srv_port;
             out << "INFO: Read server port " << srv_port << " from " << argv[ARG_COMM_IN] <<endl; 
             infile.close();
          }
          else
          {
             out << "ERROR: Can't open communications file for reading." << endl;
             return 1;
          }
   }
   else{
     cin >> srv_port;
   }

   if(srv_port <= 0)
   {
      out << "ERROR: incorrect port number." << endl;
      return 1;
   }
   serv_addr.sin_port = htons(srv_port);

   if (UDT::ERROR == UDT::connect(fhandle, (sockaddr*)&serv_addr, sizeof(serv_addr)))
   {
      out << "ERROR: connect: " << UDT::getlasterror().getErrorMessage() << endl;
      return 1;
   }

   // get size information
   int64_t size;

   if (UDT::ERROR == UDT::recv(fhandle, (char*)&size, sizeof(int64_t), 0))
   {
      out << "ERROR: send: " << UDT::getlasterror().getErrorMessage() << endl;
      return 1;
   }
 
   bool append =  argv[ARG_MODE][0]=='A' ? true : false ;

   ios::openmode mode= append? ios::out | ios::binary | ios::ate: ios::out | ios::binary | ios::trunc;

   if (append==true)
   {
     out << "INFO: appending to target file." << endl;
   }
   else
   {
     out << "INFO: target file will be overwritten." << endl;
   }

   // open target file and receive data
   ofstream ofs(argv[ARG_LOCALFILE], mode );
   int64_t recvsize; 

   if (UDT::ERROR == (recvsize = UDT::recvfile(fhandle, ofs, 0, size)))
   {
      out << "ERROR: recvfile: " << UDT::getlasterror().getErrorMessage() << endl;
      return 1;
   }

   UDT::close(fhandle);

   ofs.close();

   return 0;
}
