#ifndef WIN32
   #include <cstdlib>
   #include <arpa/inet.h>
#else
   #include <winsock2.h>
   #include <ws2tcpip.h>
#endif
#include <fstream>
#include <iostream>
#include <udt.h>
#include <cstdlib>
#include <cstring>

using namespace std;

/**
 * UNICORE UDT sendfile application
 *
 * usage: sendfile client_ip client_port local_filename [comm_file_name]
 *
 * Performs UDP hole punching and sends the specified file
 *
 * Authors: Thomas Oistrez, Bernd Schuller (JSC, Forschungszentrum Jülich GmbH)
 */
int main(int argc, char* argv[])
{
   //usage: sendfile client_ip client_port local_filename [comm_file_name]
   if (argc < 5 )
   {
      cerr << "usage: sendfile client_ip client_port local_filename [comm_file_name] " << endl;
      return 1;
   }

   bool haveCommFile=false;
   if(argc>4)haveCommFile=true;
 
   //the UDT listening socket
   UDTSOCKET serv = UDT::socket(AF_INET, SOCK_STREAM, 0);

#ifdef WIN32
   int mss = 1052;
   UDT::setsockopt(serv, 0, UDT_MSS, &mss, sizeof(int));
#endif

   //perform UDP Hole Punching by sending a dummy UDP packet to the client
   int temp_socket;
   sockaddr_in temp_addr;
   socklen_t temp_addr_size = sizeof(temp_addr);
   temp_addr.sin_family = AF_INET;
   temp_addr.sin_port = 0;
   temp_addr.sin_addr.s_addr = INADDR_ANY;
   memset(&(temp_addr.sin_zero), '\0', 8);

   temp_socket = socket(AF_INET,SOCK_DGRAM,0);
   if(temp_socket == -1)
   {
      cerr << "ERROR: socket() failed" << endl;
      return 1;
   }

   if(bind(temp_socket,(sockaddr*)&temp_addr,temp_addr_size) == -1)
   {
      cerr << "ERROR: bind() failed" << endl;
      return 1;
   }

   getsockname(temp_socket,(sockaddr*)&temp_addr,&temp_addr_size);
   int server_port = temp_addr.sin_port;

   temp_addr.sin_port = htons(atoi(argv[2]));
#ifndef WIN32
   if (inet_pton(AF_INET, argv[1], &temp_addr.sin_addr) <= 0)
#else
   if (INADDR_NONE == (temp_addr.sin_addr.s_addr = inet_addr(argv[1])))
#endif
   {
      cerr << "ERROR: incorrect network address.\n";
      return 1;
   }

   char dummy = 0;
   sendto(temp_socket,&dummy,1,0,(const struct sockaddr*)&temp_addr,sizeof(temp_addr));

   close(temp_socket);
   //end of UDP Hole Punching


   //now create UDT listening socket and wait for incoming connection
   sockaddr_in my_addr;
   int my_addr_size = sizeof(my_addr);
   my_addr.sin_family = AF_INET;
   my_addr.sin_port = server_port;
   my_addr.sin_addr.s_addr = INADDR_ANY;
   memset(&(my_addr.sin_zero), '\0', 8);

   //bind socket to any port
   if (UDT::ERROR == UDT::bind(serv, (sockaddr*)&my_addr, my_addr_size))
   {
      cerr << "ERROR: bind: " << UDT::getlasterror().getErrorMessage() << endl;
      return 1;
   }
   //find out the used port
   if(UDT::ERROR == UDT::getsockname(serv,(sockaddr*)&my_addr,&my_addr_size))
   {
      cerr << "ERROR: getsockname: " << UDT::getlasterror().getErrorMessage() << endl;
      return 1; 
   }

   cerr << "INFO: server is ready at port "<< ntohs(my_addr.sin_port) << endl;

   //print port number to commfile (or stdout). UNICORE will pick it up, so the client
   //can query it and establish the data connection

   if(haveCommFile)
   {
          cerr << "INFO: Writing local port "<< ntohs(my_addr.sin_port)<< " to" << argv[4] << endl;
          ofstream myfile;
          myfile.open(argv[4]);
          if(myfile.is_open())
          {
	    myfile << "OUT: " << ntohs(my_addr.sin_port) << endl;
             myfile.close();
          }
          else
          {
             cerr << "ERROR: Can't open communications file for writing." << endl;
             return 1;
          }
   }
   else
   {
     cout << ntohs(my_addr.sin_port) << endl;
   }
   cerr << "INFO: server is ready at port "<< ntohs(my_addr.sin_port) << endl;

   //start listening for one connection
   UDT::listen(serv, 1);

   sockaddr_in their_addr;
   int namelen = sizeof(their_addr);

   //the UDT socket for the accepted connection
   UDTSOCKET fhandle;
   //wait for incoming connection and accept it
   if (UDT::INVALID_SOCK == (fhandle = UDT::accept(serv, (sockaddr*)&their_addr, &namelen)))
   {
      cerr << "ERROR: accept: " << UDT::getlasterror().getErrorMessage() << endl;
      return 1;
   }
   //close the listening socket
   UDT::close(serv);

   // open the file
   ifstream ifs(argv[3], ios::in | ios::binary);

   ifs.seekg(0, ios::end);
   int64_t size = ifs.tellg();
   ifs.seekg(0, ios::beg);

   // send file size information
   if (UDT::ERROR == UDT::send(fhandle, (char*)&size, sizeof(int64_t), 0))
   {
      cerr << "ERROR: send: " << UDT::getlasterror().getErrorMessage() << endl;
      return 1;
   }

   UDT::TRACEINFO trace;
   UDT::perfmon(fhandle, &trace);

   // send the file
   if (UDT::ERROR == UDT::sendfile(fhandle, ifs, 0, size))
   {
      cerr << "ERROR: sendfile: " << UDT::getlasterror().getErrorMessage() << endl;
      return 1;
   }

   UDT::perfmon(fhandle, &trace);
   cerr << "INFO: speed = " << trace.mbpsSendRate << endl;

   UDT::close(fhandle);

   ifs.close();

   return 0;
}
