package de.fzj.unicore.uas.fts.udt;

import java.math.BigInteger;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import de.fzj.unicore.uas.fts.FileTransfer;
import de.fzj.unicore.uas.fts.FileTransferImpl;
import de.fzj.unicore.uas.fts.UASSizeResourceProperty;
import de.fzj.unicore.uas.fts.udt.UDTSession.State;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.uas.xnjs.XNJSFacade;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;
import de.fzj.unicore.wsrflite.xmlbeans.rp.ImmutableResourceProperty;
import de.fzj.unicore.xnjs.Configuration;
import eu.unicore.unicore6.filetransfer.udt.ExchangeParametersRequestDocument;
import eu.unicore.unicore6.filetransfer.udt.ExchangeParametersResponseDocument;
import eu.unicore.unicore6.filetransfer.udt.HostDocument;
import eu.unicore.unicore6.filetransfer.udt.IsSenderDocument;

/**
 * represents one side (the "server-side") of a single UDT-based file transfer<br/>
 * 
 * As resource properties, the host, the UDP port and the mode (is the 
 * source data on this server) are published.
 * 
 * @author oistrez
 * @author schuller
 */
public class UDTFileTransferImpl extends FileTransferImpl implements UDTFileTransfer {

	private static final Logger logger=LogUtil.getLogger(LogUtil.SERVICES,UDTFileTransferImpl.class);

	private static Map<String, UDTSession>sessions=new HashMap<String,UDTSession>();

	private transient UDTSession session;

	@Override
	protected synchronized Long getTransferredBytes() {
		if(!session.isSender()){
			try{
				transferredBytes = session.getLocalFileSize();
			}catch(Exception e){
				LogUtil.logException("Can't get size of local file.", e, logger);
			}
		}
		return transferredBytes;
	}

	void setTransferredBytes(long transferred){
		this.transferredBytes=transferred;
	}

	/**
	 * perform the UDT handshake(s)
	 */
	public ExchangeParametersResponseDocument initUDT(ExchangeParametersRequestDocument in) throws BaseFault{
		BigInteger port=null;
		String host=null;
		try{
			//store any info from client
			port=in.getExchangeParametersRequest().getClientPort();
			if(port!=null)session.setPeerPort(port.intValue());
			host=in.getExchangeParametersRequest().getClientHost();
			if(host!=null)session.setPeerHost(host);

			if(!isExport)performImportHandshake();
			else performExportHandshake();

			//send our connection params to client
			ExchangeParametersResponseDocument res = ExchangeParametersResponseDocument.Factory.newInstance();
			res.addNewExchangeParametersResponse();
			res.getExchangeParametersResponse().setServerHost(session.getMyHost());
			res.getExchangeParametersResponse().setServerPort(BigInteger.valueOf(session.getMyPort()));

			return res;

		}catch(UnknownHostException e){
			throw BaseFault.createFault("Server: Invalid connection parameters.",e); 
		}catch(Exception e){
			String msg="initUDT failed for input params: host=<"+host+"> port=<"+String.valueOf(port)+">";
			LogUtil.logException(msg, e, logger);
			throw BaseFault.createFault("Server: "+msg,e); 
		}
	}

	public void performImportHandshake()throws Exception{
		if(State.NEW.equals(session.getState())){
			session.startReceive();
		}
		else if(State.EXCHANGE_PARAMETERS.equals(session.getState())){
			session.finishReceive();
		}
		else throw new IllegalStateException("UDT session is in inconsistent state.");
	}

	public void performExportHandshake()throws Exception{
		if(State.NEW.equals(session.getState())){
			session.doSendHandshake();
		}
		else throw new IllegalStateException("UDT session is in inconsistent state.");
	}

	@Override
	public void initialise(String serviceName, Map<String, Object> map) throws Exception {
		super.initialise(serviceName, map);
		//create new UDT Transfer
		session=new UDTSession();
		//set path to local file
		String path=source!=null?source:target;
		session.setLocalFile(path);
		//set send/receive mode
		session.setSendMode(isExport);
		if(overWrite!=null && overWrite.booleanValue()==false){
			session.setAppendToFile(true);
		}
		session.setClient(getClient());
		Configuration config=XNJSFacade.get(xnjsReference, kernel).getConfiguration();
		session.setConfiguration(config);

		String myHost=config.getProperty(UDTSession.UDT_HOST, UDTSession.getLocalhostAddress());
		session.setMyHost(myHost);

		if(isExport){
			String exec=kernel.getProperty(UDTSession.UDT_SEND);
			session.setExecutable(exec);
		}
		else{
			String exec=kernel.getProperty(UDTSession.UDT_RECEIVE);
			session.setExecutable(exec);
		}
		sessions.put(getUniqueID(),session);

		HostDocument host=HostDocument.Factory.newInstance();
		host.setHost(myHost);
		ImmutableResourceProperty imr=new ImmutableResourceProperty(host);
		properties.put(UDTFileTransfer.UDT_HOSTNAME,imr);

		IsSenderDocument isSender=IsSenderDocument.Factory.newInstance();
		isSender.setIsSender(isExport);
		ImmutableResourceProperty imr2=new ImmutableResourceProperty(isSender);
		properties.put(UDTFileTransfer.UDT_IS_SENDER,imr2);

		UDTTransferredBytesProperty rp=new UDTTransferredBytesProperty(RPTransferred,this);
		properties.put(RPTransferred,rp);

		//add RP for source file size
		if(isExport){
			UASSizeResourceProperty size=new UASSizeResourceProperty(this);
			properties.put(FileTransfer.RPSize, size);
		}
	}

	@Override
	protected void customPostActivate() {
		super.customPostActivate();
		session=sessions.get(getUniqueID());
	}	

	@Override
	public void destroy(){
		if(session!=null)session.destroy();
		sessions.remove(getUniqueID());
		super.destroy();
	}

}
